from sqlalchemy import String
from sqlalchemy.ext.asyncio import AsyncAttrs
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from sqlalchemy_utils.types import UUIDType


class Base(AsyncAttrs, DeclarativeBase):
    """Base model"""

    pass


class ContextInformation(Base):
    """Contains information about the context"""

    __tablename__ = "context_information"

    id: Mapped[int] = mapped_column(primary_key=True)

    context_id: Mapped[bytes] = mapped_column(UUIDType(binary=True))
    """The id (uuid as bytes)"""

    object_id: Mapped[str] = mapped_column(String(256))
    """The object in this context"""

    parent_id: Mapped[str | None] = mapped_column(String(256), default=None)
    """id of the post, that was replied to"""



import logging

from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession

from cattle_grid.database.account import ActorForAccount
from cattle_grid.database.activity_pub_actor import Actor
from cattle_grid.model.exchange_update_actor import (
    UpdateAction,
    UpdateActionType,
    UpdateIdentifierAction,
    RenameActorAction,
    UpdatePropertyValueAction,
    UpdateUrlAction,
)

from .property_value import handle_update_property_value, handle_delete_property_value
from .identifiers import (
    handle_create_identifier,
    handle_update_identifier,
    handle_add_identifier,
)
from .urls import add_url, remove_url

logger = logging.getLogger(__name__)


async def handle_rename_actor(
    session: AsyncSession, actor: Actor, action: RenameActorAction
) -> None:
    actor_for_account = await session.scalar(
        select(ActorForAccount).where(ActorForAccount.actor == actor.actor_id)
    )
    if not actor_for_account:
        logger.warning("ActorForAccount does not exist for actor %s", actor.actor_id)
        return

    actor_for_account.name = action.name
    await session.commit()


async def handle_actor_action(
    actor: Actor, session: AsyncSession, action: UpdateAction
) -> bool:
    """Handles individual actions from the UpdateActor
    method"""
    match action.action:
        case UpdateActionType.add_identifier:
            action = UpdateIdentifierAction.model_validate(action.model_dump())
            await handle_add_identifier(actor, session, action)
            return True
        case UpdateActionType.create_identifier:
            action = UpdateIdentifierAction.model_validate(action.model_dump())
            await handle_create_identifier(actor, session, action)
            return True
        case UpdateActionType.update_identifier:
            action = UpdateIdentifierAction.model_validate(action.model_dump())
            await handle_update_identifier(actor, session, action)
            return True
        case UpdateActionType.rename:
            rename_action = RenameActorAction.model_validate(action.model_dump())
            await handle_rename_actor(session, actor, rename_action)
            return False
        case UpdateActionType.update_property_value:
            cast_action = UpdatePropertyValueAction.model_validate(action.model_dump())
            handle_update_property_value(actor, cast_action)

            return True
        case UpdateActionType.remove_property_value:
            cast_action = UpdatePropertyValueAction.model_validate(action.model_dump())
            handle_delete_property_value(actor, cast_action)

            return True

        case UpdateActionType.add_url:
            cast_action = UpdateUrlAction.model_validate(action.model_dump())
            add_url(actor, cast_action)
            return True
        case UpdateActionType.remove_url:
            cast_action = UpdateUrlAction.model_validate(action.model_dump())
            remove_url(actor, cast_action)
            return True

    return False



import pytest
import click

from unittest.mock import MagicMock

from click.testing import CliRunner

from .block_cli import add_block_command


@pytest.fixture
def config():
    settings = MagicMock()
    return settings


@pytest.fixture
def cli(tmp_path, config):
    @click.group()
    @click.pass_context
    def main(ctx):
        ctx.ensure_object(dict)
        ctx.obj["config"] = config
        ctx.obj["config_file"] = str(tmp_path / "config.toml")
        ctx.obj["config_file"] = str(tmp_path / "config.toml")
        ctx.obj["config_file_block_list"] = str(tmp_path / "config_block_list.toml")

    add_block_command(main)

    return main


def test_handle_no_config():
    @click.group()
    @click.pass_context
    def main(ctx):
        ctx.ensure_object(dict)

    add_block_command(main)
    runner = CliRunner()

    result = runner.invoke(main, ["block", "empty"])
    assert result.exit_code == 1


def test_keys_command(cli):
    runner = CliRunner()
    result = runner.invoke(cli, ["block", "empty"])

    assert result.exit_code == 0


def test_block_list_command(config, cli):
    config.domain_blocks = {"bad.example", "worse.example"}
    runner = CliRunner()
    result = runner.invoke(cli, ["block", "list"])

    assert result.exit_code == 0

    assert (
        result.stdout
        == """bad.example
worse.example
"""
    )



from unittest.mock import MagicMock

import pytest

from .config import HtmlDisplayConfiguration
from .publisher import Publisher


@pytest.fixture
def publisher():
    return Publisher(
        actor=MagicMock(actor="http://actor.test/", name="name"),
        config=HtmlDisplayConfiguration(),
    )


def test_create_html_link(publisher):
    result = publisher.create_html_link()

    assert isinstance(result, dict)


def test_update_object(publisher):
    obj = {}
    publisher.update_object(obj)

    assert isinstance(obj.get("id"), str)
    assert isinstance(obj.get("url"), list)



from fastapi import HTTPException
import pytest

from cattle_grid.testing.fixtures import *  # noqa

from .validate import validate_actor, validate_request


def test_validate_actor_none():
    with pytest.raises(HTTPException):
        validate_actor(None)


async def test_validate_request(actor_for_test, sql_session_for_test):
    with pytest.raises(HTTPException):
        await validate_request(sql_session_for_test, actor_for_test, None)



import random
import string
import uuid


from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession

from cattle_grid.extensions.examples.html_display.models import PublishingActor


def new_name():
    return "".join(random.choices(string.ascii_letters, k=12))


async def publishing_actor_for_actor_id(session: AsyncSession, actor_id: str):
    actor = await session.scalar(
        select(PublishingActor).where(PublishingActor.actor == actor_id)
    )
    if actor:
        return actor

    actor = PublishingActor(actor=actor_id, name=new_name())
    session.add(actor)
    return actor


async def object_for_name_and_uuid(session: AsyncSession, name: str, uuid: uuid.UUID):
    return None



from pydantic import BaseModel, Field


class RegistrationType(BaseModel):
    """Configuration for one registration path"""

    name: str = Field(
        examples=["dev"],
        description="name of the registration. Will be part of the path, i.e. `/register/{name}`",
    )

    permissions: list[str] = Field(
        examples=["admin"],
        description="List of permissions given to the registering account.",
    )

    extra_parameters: list[str] = Field(
        default=[],
        description="Extra parameters that should be in the request, will be stored in the actors meta information",
        examples=["fediverse"],
    )


class RegisterConfiguration(BaseModel):
    """Configuration for the register endpoint"""

    registration_types: list[RegistrationType] = Field(
        examples=[
            RegistrationType(name="dev", permissions=["dev"]),
        ],
        description="List of registration types",
    )



import pytest

from cattle_grid.testing.fixtures import *  # noqa

from .account import AccountManager


async def test_creation_not_found(sql_session_for_test):
    with pytest.raises(ValueError):
        await AccountManager.for_name_and_password(sql_session_for_test, "not", "found")


async def test_creation(sql_session_for_test, account_for_test):
    account_manager = await AccountManager.for_name_and_password(
        sql_session_for_test, account_for_test.name, account_for_test.name
    )
    assert account_manager.account == account_for_test


async def test_creation_for_name(sql_session_for_test, account_for_test):
    account_manager = await AccountManager.for_name(
        sql_session_for_test, account_for_test.name
    )
    assert account_manager.account == account_for_test



from fastapi import FastAPI

async_api_components: list[str] = ["exchange", "ap", "account"]
fastapi_components: list[str] = ["auth", "ap", "account", "rabbit", "app"]

filenames_for_component: dict[str, str] = {
    name: f"docs/assets/schemas/asyncapi_{name}.json" for name in async_api_components
}

filenames_for_openapi_components: dict[str, str] = {
    name: f"docs/assets/schemas/openapi_{name}.json"
    for name in fastapi_components
    if name != "app"
} | {"app": "docs/assets/schemas/openapi.json"}


def async_api_schema_for_component(component: str) -> str:
    if component == "ap":
        from cattle_grid.activity_pub import get_async_api_schema
    elif component == "account":
        from cattle_grid.account.router.schema import get_async_api_schema
    else:
        from cattle_grid.exchange import get_async_api_schema

    schema = get_async_api_schema().to_json()

    return schema


def fastapi_for_component(component: str) -> FastAPI:
    match component:
        case "auth":
            from cattle_grid.auth import create_app

            app = create_app()
        case "ap":
            from cattle_grid.activity_pub import get_fastapi_app

            app = get_fastapi_app()
        case "account":
            from cattle_grid.account.server.app import app
        case "rabbit":
            from cattle_grid.account.rabbit import app_for_schema

            app = app_for_schema()
        case _:
            from cattle_grid import create_app

            app = create_app()

    return app



from typing import Dict, Any

from bovine.activitystreams.utils import id_for_object


def actor_deletes_themselves(activity: Dict[str, Any]) -> bool:
    """
    Checks if activity is self delete of actor

    ```pycon
    >>> actor_deletes_themselves({"type": "Delete",
    ...     "actor": "http://actor.test/",
    ...     "object": "http://actor.test/"})
    True

    >>> actor_deletes_themselves({"type": "Delete",
    ...     "actor": "http://actor.test/",
    ...     "object": "http://other.test/"})
    False

    ```
    """

    activity_type = activity.get("type")
    if activity_type != "Delete":
        return False

    actor_id = activity.get("actor")
    object_id = id_for_object(activity.get("object"))

    if actor_id is None or object_id is None:
        return False

    return actor_id == object_id



from sqlalchemy.ext.asyncio import AsyncSession

from cattle_grid.model.account import (
    NameAndVersion,
    InformationResponse,
)
from cattle_grid.model.extension import MethodInformationModel
from cattle_grid.version import __version__

from cattle_grid.database.account import Account
from cattle_grid.manage import AccountManager


def protocol_and_backend():
    protocol = NameAndVersion(name="CattleDrive", version="0.1.0")
    backend = NameAndVersion(name="cattle_grid", version=__version__)

    return dict(protocol=protocol, backend=backend)


async def create_information_response(
    session: AsyncSession,
    account: Account,
    method_information: list[MethodInformationModel],
) -> InformationResponse:
    manager = AccountManager(account=account, session=session)

    base_urls = await manager.allowed_base_urls()

    await session.refresh(account, attribute_names=["actors"])

    return InformationResponse(
        account_name=account.name,
        base_urls=base_urls,
        actors=manager.account_information(),
        **protocol_and_backend(),
        method_information=method_information,
    )



from typing import Annotated
from fastapi import Header
from fastapi.responses import JSONResponse
from pydantic import BaseModel, Field


class ActivityResponse(JSONResponse):
    """Response that ensures the content-type is
    "application/activity+json"
    """

    media_type = "application/activity+json"


class APHeaders(BaseModel):
    """Headers every request should have. These should be added by the remote proxy."""

    x_cattle_grid_requester: str | None = Field(
        default=None, description="URI of the actor making the request"
    )
    x_cattle_grid_should_serve: str | None = Field(
        default=None, description="Type of content cattle_grid should serve"
    )
    x_ap_location: str = Field(description="URI of the resource being retrieved")


ActivityPubHeaders = Annotated[APHeaders, Header()]
"""Annotation to evaluate the [APHeaders][cattle_grid.tools.fastapi.APHeaders]"""



from typing import Any, Dict

from pydantic import BaseModel, ConfigDict, Field

from .common import WithActor


class ActivityMessage(WithActor):
    """
    Message that contains an Activity. Activity is used as the name for the 'data object' being exchanged, as is common in the Fediverse
    """

    model_config = ConfigDict(
        extra="forbid",
    )

    data: Dict[str, Any] = Field(description="The activity")


class SharedInboxMessage(BaseModel):
    """
    Message that contains an Activity. In difference to the ActivityMessage this message does not have an actor, and thus its recipients will be determined by cattle_grid.
    """

    model_config = ConfigDict(
        extra="forbid",
    )
    data: Dict[str, Any] = Field(description="Activity")


class FetchMessage(WithActor):
    """
    Used to request an ActivityPub object to be retrieved
    """

    model_config = ConfigDict(
        extra="forbid",
    )

    uri: str = Field(description="URI of the object being retrieved")



import asyncio
import click
from sqlalchemy import delete

from cattle_grid.database import database_session

from cattle_grid.database.auth import RemoteIdentity


async def prune_keys(config):
    try:
        async with database_session(config.db_uri) as session:
            await session.execute(delete(RemoteIdentity))
    except Exception:
        print("Failed to purge remote identifiers")


def add_keys_command(main):
    @main.group()
    @click.pass_context
    def keys(ctx):
        """Allows the management of public keys"""
        if "config" not in ctx.obj:
            print("Could not load config and it is necessary for keys management")
            exit(1)

    @keys.command()  # type:ignore
    @click.option(
        "--all_flag", help="Remove all known public keys", default=False, is_flag=True
    )
    @click.pass_context
    def clear(ctx, all_flag):
        try:
            if all_flag:
                asyncio.run(prune_keys(ctx.obj["config"]))
            else:
                click.echo(
                    "Currently only remove all keys is supported. If you want to do this run with the --all_flag"
                )
        except Exception as e:
            print(e)



import jinja2
from fastapi.templating import Jinja2Templates

from fastapi import APIRouter, HTTPException, Request
from fastapi.responses import HTMLResponse
from sqlalchemy import select

from cattle_grid.dependencies.fastapi import SqlSession
from cattle_grid.extensions.examples.html_display.format import format_actor_profile
from cattle_grid.extensions.examples.html_display.models import PublishedObject
from cattle_grid.tools.fastapi import ActivityResponse

from .fastapi_dependencies import (
    ActorProfile,
    PublishedObjectForUUID,
    PublishingActorForName,
)

templates = Jinja2Templates(
    env=jinja2.Environment(auto_reload=True, loader=jinja2.PackageLoader(__name__)),
)

router = APIRouter()


@router.get("/")
async def get_index():
    return "html extension"


@router.get("/object/{uuid}", response_class=ActivityResponse)
async def get__object(obj: PublishedObjectForUUID):
    return obj.data


@router.get("/html/{actor_name}", response_class=HTMLResponse)
@router.get("/html/{actor_name}/", response_class=HTMLResponse)
async def get_actor_html(
    actor: PublishingActorForName,
    profile: ActorProfile,
    request: Request,
    session: SqlSession,
):
    published_objects = await session.scalars(
        select(PublishedObject)
        .where(PublishedObject.actor == actor.actor)
        .order_by(PublishedObject.create_date.desc())
        .limit(10)
    )

    posts = [
        {
            "body": x.data.get("content"),
            "date": x.data.get("published"),
            "id": str(x.id),
        }
        for x in published_objects
    ]

    return templates.TemplateResponse(
        request,
        name="index.html.j2",
        context={
            "name": actor.name,
            "profile": format_actor_profile(profile),
            "posts": posts,
        },
    )


@router.get("/html/{actor_name}/{uuid}", response_class=HTMLResponse)
async def get_object_html(
    actor: PublishingActorForName,
    obj: PublishedObjectForUUID,
    request: Request,
):
    if actor.actor != obj.actor:
        raise HTTPException(404)

    return templates.TemplateResponse(
        request,
        name="object.html",
        context={"name": actor.name, "content": obj.data["content"]},
    )



from fast_depends import Depends
from typing import Annotated
from redis import asyncio as redis

from .config import CacheConfiguration

cache: redis.Redis | None = None

config: CacheConfiguration | None = None


def get_cache():
    global cache
    return cache


def get_config():
    global config
    return config


def set_config(cfg):
    global config
    config = cfg


CacheClient = Annotated[redis.Redis, Depends(get_cache)]
"""The cache client"""


CacheConfig = Annotated[CacheConfiguration, Depends(get_config)]
"""The configuration object passed when instantiating the extension"""



"""
Helper class to create documentation for the API

"""

from fastapi import FastAPI

from cattle_grid.version import __version__

from . import router

tags_description = [
    {
        "name": "actor",
        "description": """These functions provide the ability
to perform actions as an actor. This means interacting with
the Fediverse.

We remind the reader here that every account can have
multiple actors. The actor is usual indicated with
the `actorId` parameter.""",
    },
    {
        "name": "account",
        "description": "endpoints that do not necessitate specifying an actor",
    },
]

app = FastAPI(
    title="Account API for cattle_grid",
    description="""
This API is used to manage accounts, create actors,
manage actors, and perform actions as an actor.

Basically, it lets you control cattle_grid without using
the asynchronous API. The asynchronous API is described
[here](https://bovine.codeberg.page/cattle_grid/cattle_drive/).
""",
    openapi_tags=tags_description,
    version=__version__,
)
app.include_router(router)



from sqlalchemy import String
from sqlalchemy.ext.asyncio import AsyncAttrs
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column


class Base(AsyncAttrs, DeclarativeBase):
    """Base model"""

    pass


class RemoteIdentity(Base):
    """Stored activity in the database"""

    __tablename__ = "cattle_grid_auth_remote_identity"
    """name of the table"""

    id: Mapped[int] = mapped_column(primary_key=True)
    """The id of the key"""

    key_id: Mapped[str] = mapped_column(String(512), unique=True)
    """The URI uniquely identifying the key"""
    controller: Mapped[str] = mapped_column(String(512))
    """The URI of te controller"""
    public_key: Mapped[str] = mapped_column(String(1024))
    """The public key"""



from cattle_grid.testing.features import (
    before_all,  # noqa
    before_scenario,  # noqa
    after_scenario,  # noqa
)
from cattle_grid.testing.features.reporting import (
    before_step,  # noqa
)



from faststream.rabbit import RabbitBroker
from faststream import ExceptionMiddleware, Context

import logging
import json

from cattle_grid.account.account import account_for_actor
from cattle_grid.model.account import ErrorMessage
from cattle_grid.dependencies import CorrelationId, AccountExchange, SqlSession
from cattle_grid.dependencies.processing import RoutingKey

logger = logging.getLogger(__name__)

exception_middleware = ExceptionMiddleware()


@exception_middleware.add_handler(Exception)
async def exception_handler(
    exception: Exception,
    routing_key: RoutingKey,
    correlation_id: CorrelationId,
    session: SqlSession,
    account_exchange: AccountExchange,
    message=Context("message.body"),
    broker: RabbitBroker = Context(),
):
    try:
        data = json.loads(message)
        actor_id = data["actor"]
    except Exception:
        logger.exception(exception)
        logger.exception(message)
        return

    logger.info("Exception for %s", actor_id)

    account = await account_for_actor(session, actor_id)
    if not account:
        return
    name = account.name

    logger.error("Processing error occurred in exchange for account %s", name)

    await broker.publish(
        ErrorMessage(message=str(exception), routing_key=routing_key),
        routing_key=f"error.{name}",
        exchange=account_exchange,
        correlation_id=correlation_id,
    )



from fastapi import HTTPException
import pytest


from .testing import *  # noqa
from .fastapi_dependencies import publishing_actor_for_name


async def test_publishing_actor_for_name(sql_session_for_test):
    with pytest.raises(HTTPException):
        await publishing_actor_for_name(sql_session_for_test, "name")



import logging

from sqlalchemy.ext.asyncio import AsyncSession

from cattle_grid.account.account import AccountAlreadyExists, create_account
from cattle_grid.dependencies.globals import global_container

from cattle_grid.activity_pub.actor import create_actor
from cattle_grid.database.account import ActorForAccount

logger = logging.getLogger(__name__)


async def create_test_accounts(session: AsyncSession):
    config = global_container.get_config().testing

    if not config.enable:  # type:ignore
        return

    logger.warning("running in testing mode")
    logger.warning(
        "\n   __            __  _            \n  / /____  _____/ /_(_)___  ____ _\n / __/ _ \\/ ___/ __/ / __ \\/ __ `/\n/ /_/  __(__  ) /_/ / / / / /_/ / \n\\__/\\___/____/\\__/_/_/ /_/\\__, /  \n                         /____/   \n"
    )
    accounts: list[dict] = config.accounts  # type:ignore

    for info in accounts:
        try:
            account = await create_account(
                session,
                info["name"],
                info["password"],
                permissions=info.get("permissions", []),
                meta_information=info.get(
                    "meta_information", {"testing": "created by testing"}
                ),
            )
            logger.info("created account %s", info["name"])

            actors = info.get("actors", [])
            for actor_info in actors:
                actor = await create_actor(
                    session,
                    actor_info.get("base_url"),
                    preferred_username=actor_info.get("handle"),
                )
                actor.automatically_accept_followers = True
                session.add(
                    ActorForAccount(
                        account=account,
                        actor=actor.actor_id,
                        name=actor_info.get("name"),
                    )
                )

                await session.commit()

        except AccountAlreadyExists:
            pass



import uuid6
from urllib.parse import urlparse

from pydantic import BaseModel, Field


def determine_url_start(actor_id, prefix):
    """
    Used to determine the url of a stored object

    ```pycon
    >>> determine_url_start("http://abel.example/actor/alice",
    ...     "/simple/storage/")
    'http://abel.example/simple/storage/'

    ```
    """
    parsed = urlparse(actor_id)

    return f"{parsed.scheme}://{parsed.netloc}{prefix}"


class SimpleStorageConfiguration(BaseModel):
    """Configuration of the simple storage extension"""

    prefix: str = Field(
        "/simple_storage/",
        description="Path to use before the generated uuid. The protocol and domain will be extracted from the actor id. See [cattle_grid.extensions.examples.simple_storage.config.determine_url_start][].",
    )

    def url_start(self, actor_id):
        return determine_url_start(actor_id, self.prefix)

    def make_id(self, actor_id):
        new_uuid = uuid6.uuid7()
        return self.url_start(actor_id) + str(new_uuid), new_uuid



from unittest.mock import AsyncMock

from sqlalchemy import select
from cattle_grid.testing.fixtures import *  # noqa

from cattle_grid.database.activity_pub_actor import StoredActivity
from cattle_grid.model.processing import StoreActivityMessage

from .store_activity import store_activity_subscriber


async def test_store_activity(sql_session_for_test, actor_for_test):
    broker = AsyncMock()
    activity = {
        "actor": actor_for_test.actor_id,
        "type": "Activity",
        "to": ["http://remote.test/actor"],
    }
    msg = StoreActivityMessage(actor=actor_for_test.actor_id, data=activity)

    await store_activity_subscriber(
        msg, actor_for_test, session=sql_session_for_test, broker=broker
    )

    assert 1 == len(
        [x for x in await sql_session_for_test.scalars(select(StoredActivity))]
    )

    broker.publish.assert_awaited_once()



from cattle_grid.model.lookup import Lookup
from cattle_grid.extensions import Extension

from .lookup import ordered_lookups


async def lookup_1(lookup: Lookup) -> Lookup:
    return lookup


async def lookup_2(lookup: Lookup) -> Lookup:
    return lookup


async def lookup_3(lookup: Lookup) -> Lookup:
    return lookup


def test_ordered_lookups():
    one = Extension(name="one", lookup_method=lookup_1, lookup_order=0, module=__name__)
    two = Extension(name="two", lookup_method=lookup_2, lookup_order=2, module=__name__)
    three = Extension(
        name="three", lookup_method=lookup_3, lookup_order=1, module=__name__
    )

    result = ordered_lookups([one, two, three])

    assert result == [lookup_2, lookup_3, lookup_1]


def test_ordered_lookups_filters():
    one = Extension(name="one", module=__name__)

    result = ordered_lookups([one])

    assert len(result) == 0



import logging

from bovine import BovineActor
from sqlalchemy import select
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.asyncio import AsyncSession

from cattle_grid.database.activity_pub import InboxLocation
from cattle_grid.model import FetchMessage

from cattle_grid.dependencies import LookupAnnotation, SqlSession
from cattle_grid.model.lookup import Lookup

from cattle_grid.model.processing import ToSendMessage
from .common import MessageBovineActor


logger = logging.getLogger(__name__)


async def fetch_object(
    message: FetchMessage,
    actor: MessageBovineActor,
    lookup: LookupAnnotation,
):
    """Handles retrieving a remote object"""

    try:
        lookup_result = await lookup(Lookup(uri=message.uri, actor=message.actor))
        if lookup_result.result:
            return lookup_result.result
    except Exception as e:
        logger.error("Something went up with lookup")
        logger.exception(e)

        lookup_result = Lookup(uri=message.uri, actor=message.actor)

    result = await actor.get(lookup_result.uri, fail_silently=True)

    return result


async def resolve_inbox(session: AsyncSession, actor: BovineActor, target: str):
    """Resolves the inbox of target for actor using
    a cache"""
    cached = await session.scalar(
        select(InboxLocation).where(InboxLocation.actor == target)
    )

    if cached:
        return cached.inbox

    target_actor = await actor.get(target)
    if not target_actor:
        return None

    inbox = target_actor.get("inbox")
    if inbox is None:
        return

    session.add(InboxLocation(actor=target, inbox=inbox))

    try:
        await session.commit()
    except IntegrityError:
        ...
    return inbox


async def sending_message(
    message: ToSendMessage, actor: MessageBovineActor, sql_session: SqlSession
):
    """Handles sending a message"""
    inbox = await resolve_inbox(sql_session, actor, message.target)
    if inbox:
        result = await actor.post(inbox, message.data)
        logger.info("Got %s for sending to %s", str(result), inbox)



import json
import logging

from cattle_grid.model.lookup import Lookup

from .dependencies import CacheClient

logger = logging.getLogger(__name__)


async def cache_lookup(lookup: Lookup, cache: CacheClient) -> Lookup:
    cache_result = await cache.get(lookup.uri)

    logger.info("Got result from cache %s", cache_result)

    if not cache_result:
        return lookup

    lookup.result = json.loads(cache_result)

    logger.info(lookup)

    return lookup



from .user import *  # noqa
from .follow import *  # noqa
from .collection import *  # noqa
from .messaging import *  # noqa
from .block import *  # noqa
from .account import *  # noqa
from .message_content import *  # noqa
from .drive import *  # noqa
from .http import *  # noqa



from behave import given, then
from cattle_grid.model.exchange import UpdateActorMessage
from cattle_grid.model.exchange_update_actor import (
    UpdateActionType,
    UpdateIdentifierAction,
)
from cattle_grid.testing.features import publish_as


@given('"{alice}" adds "{identifier}" as a primary identifier')  # type: ignore
async def add_identifier(context, alice, identifier):
    actor = context.actors[alice]

    msg = UpdateActorMessage(
        actor=actor.get("id"),
        actions=[
            UpdateIdentifierAction(
                action=UpdateActionType.create_identifier,
                identifier=identifier,
                primary=True,
            )
        ],
    ).model_dump()

    await publish_as(context, alice, "update_actor", msg)


@then('The preferred username is "{alex}"')  # type: ignore
def check_preferred_username(context, alex):
    assert context.result.get("preferredUsername") == alex


@then('"{identifier}" is contained in the identifiers array')  # type: ignore
def check_identifiers(context, identifier):
    assert identifier in context.result.get("identifiers")


@then('"{identifier}" is not contained in the identifiers array')  # type: ignore
def check_not_in_identifiers(context, identifier):
    assert identifier not in context.result.get("identifiers")



import pytest

from urllib.parse import urlparse
from fastapi import FastAPI
from fastapi.testclient import TestClient
from cattle_grid.activity_pub.actor import create_actor

from cattle_grid.testing.fixtures import *  # noqa

from . import router


@pytest.fixture
def test_client():
    app = FastAPI()
    app.include_router(router)

    yield TestClient(app)


async def test_webfinger_plain(test_client):
    response = test_client.get("/ap/.well-known/webfinger")

    assert response.status_code == 422


async def test_webfinger_not_found(test_client):
    response = test_client.get(
        "/ap/.well-known/webfinger", params={"resource": "acct:unknown@nowhere"}
    )

    assert response.status_code == 404


async def test_webfinger_found(test_client, sql_session_for_test):
    identifier = "acct:me@localhost"

    await create_actor(
        sql_session_for_test,
        "http://localhost/ap/",
        identifiers={"webfinger": identifier},
    )

    response = test_client.get(
        "/ap/.well-known/webfinger", params={"resource": identifier}
    )

    assert response.status_code == 200


async def test_nodeinfo(test_client):
    response = test_client.get(
        "/ap/.well-known/nodeinfo",
        headers={"x-ap-location": "http://testserver/.well-known/nodeinfo"},
    )
    assert response.status_code == 200

    data = response.json()
    assert len(data["links"]) == 1

    link = data["links"][0]

    assert link["type"] == "http://nodeinfo.diaspora.software/ns/schema/2.0"
    assert link["href"] == "http://testserver/.well-known/nodeinfo_2.0"

    node_info_response = test_client.get("/ap" + urlparse(link["href"]).path)
    assert node_info_response.status_code == 200

    assert node_info_response.headers["content-type"] == "application/jrd+json"



from enum import StrEnum, auto
from sqlalchemy import JSON, ForeignKey, String
from sqlalchemy.orm import Mapped, mapped_column, relationship


from .activity_pub import Base


class ActorStatus(StrEnum):
    """Status actors can have for an account"""

    active = auto()
    deleted = auto()


class Account(Base):
    __tablename__ = "account"
    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str] = mapped_column(String(256))
    password_hash: Mapped[str] = mapped_column(String(256))
    meta_information: Mapped[dict] = mapped_column(JSON(), default={})

    actors: Mapped[list["ActorForAccount"]] = relationship(viewonly=True)
    permissions: Mapped[list["Permission"]] = relationship(viewonly=True)


class ActorForAccount(Base):
    __tablename__ = "actorforaccount"
    id: Mapped[int] = mapped_column(primary_key=True)
    account_id: Mapped[int] = mapped_column(
        ForeignKey("account.id", ondelete="CASCADE")
    )
    account: Mapped[Account] = relationship(lazy="joined")

    actor: Mapped[str] = mapped_column(String(256))
    name: Mapped[str] = mapped_column(
        String(256),
        default="NO NAME",
    )
    status: Mapped[ActorStatus] = mapped_column(String(10), default=ActorStatus.active)

    groups: Mapped[list["ActorGroup"]] = relationship(viewonly=True)


class AuthenticationToken(Base):
    __tablename__ = "authenticationtoken"
    token: Mapped[str] = mapped_column(String(65), primary_key=True)
    account_id: Mapped[int] = mapped_column(
        ForeignKey("account.id", ondelete="CASCADE")
    )
    account: Mapped[Account] = relationship(lazy="joined")


class Permission(Base):
    __tablename__ = "permission"
    id: Mapped[int] = mapped_column(primary_key=True)
    account_id: Mapped[int] = mapped_column(
        ForeignKey("account.id", ondelete="CASCADE")
    )
    account: Mapped[Account] = relationship(lazy="joined")

    name: Mapped[str] = mapped_column(String(256))


class ActorGroup(Base):
    __tablename__ = "actorgroup"
    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str] = mapped_column(String(256))
    actor_id: Mapped[int] = mapped_column(
        ForeignKey("actorforaccount.id", ondelete="CASCADE")
    )
    actor: Mapped[ActorForAccount] = relationship(lazy="joined")



import logging

from typing import Annotated
from faststream import Context
from faststream.rabbit import RabbitBroker
from sqlalchemy import select
from sqlalchemy.orm.attributes import flag_modified

from cattle_grid.activity_pub.actor import (
    update_for_actor_profile,
    delete_actor,
    delete_for_actor_profile,
)
from cattle_grid.database.account import ActorForAccount, ActorStatus
from cattle_grid.dependencies import SqlSession, InternalExchange
from cattle_grid.dependencies.globals import global_container
from cattle_grid.dependencies.processing import MessageActor

from cattle_grid.model.exchange import (
    UpdateActorMessage,
    DeleteActorMessage,
)
from cattle_grid.model.processing import StoreActivityMessage

from .actor_update import handle_actor_action

logger = logging.getLogger(__name__)


async def update_actor(
    message: UpdateActorMessage,
    actor: MessageActor,
    session: SqlSession,
    broker: RabbitBroker = Context(),
) -> None:
    """Should be used asynchronously"""
    send_update = False

    for action in message.actions:
        try:
            if await handle_actor_action(actor, session, action):
                await session.commit()
                await session.refresh(actor, attribute_names=["identifiers"])
                send_update = True
        except Exception as e:
            logger.error(
                "Something went wrong when handling action of type %s",
                action.action.value,
            )
            logger.exception(e)

            # FIXME publish to error? How?

    if message.profile:
        actor.profile.update(message.profile)
        flag_modified(actor, "profile")

        logger.info("Updating actor %s", actor.actor_id)
        await session.commit()

        send_update = True

    if message.autoFollow is not None:
        actor.automatically_accept_followers = message.autoFollow
        await session.commit()

    await session.refresh(actor)

    if send_update:
        await broker.publish(
            StoreActivityMessage(
                actor=message.actor, data=update_for_actor_profile(actor)
            ),
            routing_key="store_activity",
            exchange=global_container.internal_exchange,
        )


async def delete_actor_handler(
    message: DeleteActorMessage,
    actor: MessageActor,
    session: SqlSession,
    internal_exchange: InternalExchange,
    broker: Annotated[RabbitBroker, Context()],
) -> None:
    """
    Deletes the actor by id. Should be used asynchronously.
    """

    logger.info("Deleting actor %s", message.actor)
    actor_for_account = await session.scalar(
        select(ActorForAccount).where(ActorForAccount.actor == message.actor)
    )
    if actor_for_account:
        logger.info("setting account to deleted")
        actor_for_account.status = ActorStatus.deleted

    await session.refresh(actor, attribute_names=["identifiers"])

    await broker.publish(
        StoreActivityMessage(
            actor=actor.actor_id, data=delete_for_actor_profile(actor)
        ),
        routing_key="store_activity",
        exchange=internal_exchange,
    )

    await delete_actor(session, actor)
    await session.commit()



from fast_depends import inject

from . import Extension, BaseConfig


async def test_transform():
    ext = Extension("test", module=__name__)

    @ext.transform(inputs=["inputs"], outputs=["outputs"])
    async def transformer(a: dict):
        return {"b": 1}

    assert ext.transformer_inputs == ["inputs"]
    assert ext.transformer_outputs == ["outputs"]

    assert ext.transformer

    result = await ext.transformer({"a": 1})

    assert result == {"b": 1}


async def test_extension_with_config():
    class TestConfig(BaseConfig):
        a: str

    ext = Extension("test", config_class=TestConfig, module=__name__)
    config = {"a": "value"}

    ext.configure(config)

    def test_function(cfg: ext.Config):  # type:ignore
        assert isinstance(cfg, TestConfig)
        assert cfg.a == "value"

    inject(test_function)()  # type:ignore


def test_extension_with_subscribe():
    ext = Extension("test", module=__name__)

    @ext.subscribe("test_key")
    async def subscriber(msg: dict):
        """Description"""
        pass

    assert len(ext.method_information) == 1

    info = ext.method_information[0]

    assert info.routing_key == "test_key"
    assert info.module == "cattle_grid.extensions.test_init"
    assert info.description == "Description"


def test_extension_with_subscribe_specify_description():
    ext = Extension("test", module=__name__)

    @ext.subscribe("test_key", description="other")
    async def subscriber(msg: dict):
        """Description"""
        pass

    assert len(ext.method_information) == 1

    info = ext.method_information[0]

    assert info.description == "other"


def test_extension_with_subscribe_incoming_outgoing():
    ext = Extension("test", module=__name__)

    @ext.subscribe("incoming.Create")
    async def subscriber(msg: dict):
        pass

    @ext.subscribe("outgoing.Create")
    async def outgoing_subscriber(msg: dict):
        pass

    assert len(ext.method_information) == 0


def test_extension_subscribe_wildcard_queues():
    ext = Extension("test", module=__name__)

    @ext.subscribe("incoming.*")
    async def start_subscriber(message: dict):
        pass

    @ext.subscribe("incoming.#")
    async def hash_subscriber(message: dict):
        pass

    @ext.subscribe_on_account_exchange("receive.*")
    async def account_subscriber(message: dict):
        pass



from typing import Annotated

from fast_depends import Depends
from cattle_grid.dependencies import CommittingSession
from cattle_grid.dependencies.processing import MessageActor

from .storage import publishing_actor_for_actor_id
from .models import PublishingActor as PublishingActorModel


async def publishing_actor(session: CommittingSession, actor: MessageActor):
    # FIXME: Should one at this point require the actor to be part
    # of the html_display group.
    # If yes, how should this be configured?

    return await publishing_actor_for_actor_id(session, actor.actor_id)


PublishingActor = Annotated[PublishingActorModel, Depends(publishing_actor)]
"""Returns the publishing actor"""



import pytest

from cattle_grid.testing.fixtures import *  # noqa
from cattle_grid.database.activity_pub_actor import Follower, Blocking


from . import create_actor
from .requester import is_valid_requester


@pytest.fixture
async def test_actor(sql_session_for_test):
    actor = await create_actor(sql_session_for_test, "http://localhost/ap/")

    sql_session_for_test.add_all(
        [
            Follower(
                actor=actor,
                follower="http://follower.test",
                accepted=True,
                request="xxx",
            ),
            Blocking(
                actor=actor, blocking="http://blocking.test", active=True, request="xxx"
            ),
        ]
    )
    await sql_session_for_test.commit()

    return actor


@pytest.mark.parametrize(
    "obj,expected",
    [
        ({}, False),
        ({"to": ["http://remote.example"]}, True),
        ({"to": ["as:Public"]}, True),
    ],
)
async def test_is_valid_requester_remote(
    sql_session_for_test, test_actor, obj, expected
):
    valid = await is_valid_requester(
        sql_session_for_test, "http://remote.example", test_actor, obj
    )

    assert valid == expected


@pytest.mark.parametrize(
    "obj,expected",
    [
        ({}, False),
        ({"to": ["http://blocking.test"]}, False),
        ({"to": ["as:Public"]}, False),
    ],
)
async def test_is_valid_requester_blocking(
    sql_session_for_test, test_actor, obj, expected
):
    valid = await is_valid_requester(
        sql_session_for_test, "http://blocking.test", test_actor, obj
    )

    assert valid == expected



"""The before_all, _scenario, and after_scenario functions
need to be imported in your environment.py file, e.g.


```python title="features/environment.py"
--8<-- "./features/environment.py"
```
"""

import asyncio
import aiohttp
import secrets
import logging

from almabtrieb.exceptions import ErrorMessageException

from cattle_grid.config import load_settings
from cattle_grid.account.account import delete_account
from cattle_grid.database import database_session
from .reporting import publish_reporting

logger = logging.getLogger(__name__)


def id_generator_for_actor(actor):
    def gen():
        return actor.get("id") + "/" + secrets.token_hex(8)

    return gen


async def create_session(context):
    if not context.session:
        context.session = aiohttp.ClientSession()


async def close_session(context):
    config = load_settings()

    async with database_session(db_uri=config.db_uri) as session:  # type: ignore
        for name, connection in context.connections.items():
            try:
                if name in context.actors:
                    await connection.trigger(
                        "delete_actor",
                        {
                            "actor": context.actors[name].get("id"),
                        },
                    )
                    await asyncio.sleep(0.1)

                await delete_account(session, name, name)
            except Exception as e:
                logger.exception(e)

    await context.session.close()

    try:
        for connection in context.connections.values():
            connection.task.cancel()
            try:
                await connection.task
            except asyncio.CancelledError:
                pass
    except Exception as e:
        logger.exception(e)


def before_all(context):
    """Called in features/environment.py"""
    tortoise_logger = logging.getLogger("tortoise")
    tortoise_logger.setLevel(logging.WARNING)
    context.session = None

    context.actors = {}
    context.connections = {}


def before_scenario(context, scenario):
    """Called in features/environment.py"""
    asyncio.get_event_loop().run_until_complete(create_session(context))

    context.actors = {}
    context.connections = {}

    asyncio.get_event_loop().run_until_complete(
        publish_reporting(
            "scenario",
            {
                "name": scenario.name,
                "file": scenario.filename,
                "description": scenario.description,
            },
        )
    )


def after_scenario(context, scenario):
    """Called in features/environment.py"""

    asyncio.get_event_loop().run_until_complete(
        publish_reporting(
            "scenario_end",
            {},
        )
    )

    if context.session:
        asyncio.get_event_loop().run_until_complete(close_session(context))
        context.session = None


async def publish_as(
    context, username: str, method: str, data: dict, timeout: float = 0.3
):
    """Publishes a message through the gateway

    :param data: The message to be published"""
    connection = context.connections[username]

    await connection.trigger(method, data)
    await asyncio.sleep(timeout)


async def fetch_request(context, username: str, uri: str) -> dict | None:
    """Sends a fetch request for the uri through the gateway

    :param context: The behave context
    :param username: username performing the result
    :param uri: URI being looked up
    :return:
    """
    connection = context.connections[username]
    actor = context.actors[username].get("id")

    await asyncio.sleep(0.1)

    try:
        data = await connection.fetch(actor, uri)

        assert data.uri == uri

        return data.data
    except ErrorMessageException:
        return None


def send_message_as_actor(actor, activity):
    return {"actor": actor.get("id"), "data": activity}



import uuid

from .config import SimpleStorageConfiguration


def test_make_id():
    config = SimpleStorageConfiguration(prefix="/simple/storage/")

    new_id, new_uuid = config.make_id("http://alice.example/some/path")

    expected_start = "http://alice.example/simple/storage/"

    assert new_id.startswith(expected_start)

    generated_uuid = uuid.UUID(new_id.removeprefix(expected_start))

    assert isinstance(generated_uuid, uuid.UUID)
    assert generated_uuid == new_uuid



from dynaconf import Validator

base_validators = [
    Validator("amqp_uri", default="amqp://localhost"),
    Validator("db_uri", default="sqlite:///cattle_grid.db"),
    Validator("enable_reporting", cast=bool, default=False),
    Validator("processor_in_app", cast=bool, default=False),
    Validator("permissions", default={}),
]
"""Validates the basic configuration"""

auth_validators = [
    Validator("auth.require_signature_for_activity_pub", default=True),
]
"""Validates the authentication configuration"""

activity_pub_validators = [
    Validator("activity_pub.internal_exchange", default="cattle_grid_internal"),
    Validator("activity_pub.exchange", default="cattle_grid"),
    Validator("activity_pub.account_exchange", default="amq.topic"),
]
"""Validators for ActivityPub"""

gateway_admin_validations = [
    Validator("gateway.admin.enable", cast=bool, default=False),
    Validator("gateway.admin.enable_reset", cast=bool, default=False),
]
"""Validators for the gateway"""

account_validations = [
    Validator(
        "account.forbidden_names",
        default=lambda a, b: list(["bovine", "cattle_grid", "admin", "guest"]),
        cast=list,
    ),
    Validator("account.allowed_name_regex", cast=str, default="^[a-zA-Z0-9_]{1,16}$"),
]
"""Validators for the account"""

plugins_validations = [
    Validator("plugins", default=lambda a, b: list([]), cast=list),
]
"""Validators for the plugins"""


frontend_validations = [
    Validator(
        "frontend.base_urls",
        default=lambda a, b: list([]),
        cast=lambda x: [str(y) for y in x],
        condition=lambda items: all(
            x.startswith("http://") or x.startswith("https://") for x in items
        ),
    )
]
"""Validators for the frontend"""


extensions_validations = [
    Validator("extensions", default=lambda a, b: list([]), cast=list),
]
"""Validators for the plugins"""

testing_validators = [
    Validator("testing.enable", cast=bool, default=False),
    Validator("testing.accounts", default=lambda a, b: list([]), cast=list),
]
"""Validators for testing"""

all_validators = (
    base_validators
    + activity_pub_validators
    + auth_validators
    + gateway_admin_validations
    + plugins_validations
    + frontend_validations
    + extensions_validations
    + account_validations
    + testing_validators
)



from faststream import Context
from faststream.rabbit import RabbitBroker
from sqlalchemy import select


from bovine.activitystreams.utils import recipients_for_object

from cattle_grid.model import SharedInboxMessage
from cattle_grid.database.activity_pub_actor import Actor, Following
from cattle_grid.activity_pub.enqueuer import enqueue_from_inbox
from cattle_grid.dependencies.globals import global_container
from cattle_grid.dependencies import SqlSession


def to_result_set(result):
    return set(result.all())


async def handle_shared_inbox_message(
    message: SharedInboxMessage,
    session: SqlSession,
    broker: RabbitBroker = Context(),
):
    """
    This method is used to handle incoming messages from the shared inbox.
    """

    recipients = recipients_for_object(message.data)
    sender = message.data.get("actor")

    if sender is None:
        return

    local_actor_ids = to_result_set(
        await session.scalars(
            select(Actor.actor_id).where(Actor.actor_id.in_(recipients))
        )
    )
    following_actor_ids = {
        x.actor.actor_id
        for x in await session.scalars(
            select(Following)
            .where(Following.following == sender)
            .where(Following.accepted)
        )
    }

    for actor in local_actor_ids | following_actor_ids:
        await enqueue_from_inbox(
            broker, global_container.internal_exchange, actor, message.data
        )



from dataclasses import dataclass, field
from uuid import UUID

from bovine.activitystreams.utils import as_list
from uuid6 import uuid7


from .config import HtmlDisplayConfiguration
from .models import PublishingActor


@dataclass
class Publisher:
    """Class for manipulating objects being published"""

    actor: PublishingActor
    config: HtmlDisplayConfiguration
    uuid: UUID = field(default_factory=uuid7)

    def update_object(self, obj):
        obj_id = self.config.url_start(self.actor.actor) + str(self.uuid)
        obj["id"] = obj_id
        self._add_url_to_obj(obj)

    def _add_url_to_obj(self, obj: dict):
        url_list = as_list(obj.get("url", []))

        url = (
            self.config.html_url_start(self.actor.actor)
            + self.actor.name
            + "/"
            + str(self.uuid)
        )

        obj["url"] = url_list + [self.create_html_link(url)]

    def create_html_link(self, url: str | None = None):
        if url is None:
            url = self.actor.actor_id + "#html"
        return {"type": "Link", "mediaType": "text/html", "href": url}



"""foreign keys for actor

Revision ID: 278e6319bf31
Revises: 95e4ef3e61b7
Create Date: 2025-08-29 14:28:50.091815

"""

from typing import Sequence, Union

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "278e6319bf31"
down_revision: Union[str, Sequence[str], None] = "95e4ef3e61b7"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """Upgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(op.f("blocking_actor_id_fkey"), "blocking", type_="foreignkey")
    op.create_foreign_key(
        None, "blocking", "actor", ["actor_id"], ["id"], ondelete="CASCADE"
    )
    op.drop_constraint(op.f("follower_actor_id_fkey"), "follower", type_="foreignkey")
    op.create_foreign_key(
        None, "follower", "actor", ["actor_id"], ["id"], ondelete="CASCADE"
    )
    op.drop_constraint(op.f("following_actor_id_fkey"), "following", type_="foreignkey")
    op.create_foreign_key(
        None, "following", "actor", ["actor_id"], ["id"], ondelete="CASCADE"
    )
    op.drop_constraint(
        op.f("publicidentifier_actor_id_fkey"), "publicidentifier", type_="foreignkey"
    )
    op.create_foreign_key(
        None, "publicidentifier", "actor", ["actor_id"], ["id"], ondelete="CASCADE"
    )
    op.drop_constraint(
        op.f("storedactivity_actor_id_fkey"), "storedactivity", type_="foreignkey"
    )
    op.create_foreign_key(
        None, "storedactivity", "actor", ["actor_id"], ["id"], ondelete="CASCADE"
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, "storedactivity", type_="foreignkey")
    op.create_foreign_key(
        op.f("storedactivity_actor_id_fkey"),
        "storedactivity",
        "actor",
        ["actor_id"],
        ["id"],
    )
    op.drop_constraint(None, "publicidentifier", type_="foreignkey")
    op.create_foreign_key(
        op.f("publicidentifier_actor_id_fkey"),
        "publicidentifier",
        "actor",
        ["actor_id"],
        ["id"],
    )
    op.drop_constraint(None, "following", type_="foreignkey")
    op.create_foreign_key(
        op.f("following_actor_id_fkey"), "following", "actor", ["actor_id"], ["id"]
    )
    op.drop_constraint(None, "follower", type_="foreignkey")
    op.create_foreign_key(
        op.f("follower_actor_id_fkey"), "follower", "actor", ["actor_id"], ["id"]
    )
    op.drop_constraint(None, "blocking", type_="foreignkey")
    op.create_foreign_key(
        op.f("blocking_actor_id_fkey"), "blocking", "actor", ["actor_id"], ["id"]
    )
    # ### end Alembic commands ###



"""ActivityPub related functionality"""

import logging
from fastapi import APIRouter, HTTPException

from bovine.activitystreams import OrderedCollection
from bovine.activitystreams.utils import as_list
from fastapi.responses import RedirectResponse
from sqlalchemy import select
from sqlalchemy.orm import joinedload

from cattle_grid.database.activity_pub_actor import Actor
from cattle_grid.dependencies.fastapi import SqlSession

from cattle_grid.activity_pub.actor import (
    actor_to_object,
)
from cattle_grid.activity_pub.actor.relationship import (
    followers_for_actor,
    following_for_actor,
)
from cattle_grid.tools.fastapi import ActivityResponse, ActivityPubHeaders

from .validate import validate_request

logger = logging.getLogger(__name__)

ap_router = APIRouter()


def extract_html_url(actor: Actor) -> str | None:
    urls = as_list(actor.profile.get("url", []))

    for url in urls:
        if isinstance(url, str):
            return url
        if url.get("mediaType").startswith("text/html"):
            return url.get("href")

    return None


@ap_router.get("/actor/{id_str}", response_class=ActivityResponse)
async def actor_profile(headers: ActivityPubHeaders, session: SqlSession):
    """Returns the actor"""
    logger.debug("Request for actor at %s", headers.x_ap_location)
    actor = await session.scalar(
        select(Actor)
        .where(Actor.actor_id == headers.x_ap_location)
        .options(joinedload(Actor.identifiers))
    )

    if headers.x_cattle_grid_should_serve == "html":
        if not actor:
            raise HTTPException(404)
        html_url = extract_html_url(actor)
        if html_url:
            return RedirectResponse(html_url)
        raise HTTPException(406)

    actor = await validate_request(session, actor, headers.x_cattle_grid_requester)

    result = actor_to_object(actor)
    return result


@ap_router.get("/outbox/{id_str}", response_class=ActivityResponse)
async def outbox(headers: ActivityPubHeaders, session: SqlSession):
    """Returns an empty ordered collection as outbox"""
    actor = await session.scalar(
        select(Actor).where(Actor.outbox_uri == headers.x_ap_location)
    )
    actor = await validate_request(session, actor, headers.x_cattle_grid_requester)

    return OrderedCollection(id=headers.x_ap_location).build()


@ap_router.get("/following/{id_str}", response_class=ActivityResponse)
async def following(id_str, headers: ActivityPubHeaders, session: SqlSession):
    """Returns the following"""

    actor = await session.scalar(
        select(Actor).where(Actor.following_uri == headers.x_ap_location)
    )
    actor = await validate_request(session, actor, headers.x_cattle_grid_requester)

    following = await following_for_actor(session, actor)
    return OrderedCollection(id=headers.x_ap_location, items=list(following)).build()


@ap_router.get("/followers/{id_str}", response_class=ActivityResponse)
async def followers(id_str, headers: ActivityPubHeaders, session: SqlSession):
    """Returns the followers"""
    actor = await session.scalar(
        select(Actor).where(Actor.followers_uri == headers.x_ap_location)
    )
    actor = await validate_request(session, actor, headers.x_cattle_grid_requester)

    followers = await followers_for_actor(session, actor)
    return OrderedCollection(id=headers.x_ap_location, items=list(followers)).build()



import pytest

from dynaconf import Dynaconf

from cattle_grid.testing.fixtures import auth_config_file, auth_config  # noqa

from .auth import get_auth_config, AuthNotConfigured, AuthConfig, new_auth_config
from .settings import get_settings


def test_get_auth_config_blank():
    with pytest.raises(AuthNotConfigured):
        get_auth_config(settings=Dynaconf())


def test_get_auth_blank(auth_config_file):  # noqa
    result = get_auth_config(get_settings([auth_config_file]))

    assert isinstance(result, AuthConfig)


def test_new_auth_config():
    config = new_auth_config("http://localhost/cattle_grid_actor")

    assert config.actor_acct_id.startswith("acct:")
    assert config.actor_acct_id.endswith("@localhost")



"""
Puts performing a webfinger lookup in the processing chain.
In a production setting, one can include this extension
via

```toml title="cattle_grid.toml"
[[extensions]]
module = "cattle_grid.extensions.examples.webfinger_lookup"

lookup_order = 10
```

Where the higher the `lookup_order`, the earlier the conversion from `acct` to resolvable URI happens.

In a test setting, one can make webfinger use http via

```toml title="cattle_grid.toml"
[[extensions]]
module = "cattle_grid.extensions.examples.webfinger_lookup"

config = { protocol = "http" }
```

"""

from pydantic import BaseModel, Field
from enum import StrEnum, auto

from bovine.clients import lookup_uri_with_webfinger

from cattle_grid.model.lookup import Lookup
from cattle_grid.extensions import Extension
from cattle_grid.dependencies import ClientSession


class Protocol(StrEnum):
    """Protocol to use for webfinger lookups"""

    http = auto()
    https = auto()


class WebfingerLookupConfig(BaseModel):
    """Configuration for the webfinger lookup"""

    protocol: Protocol = Field(
        default=Protocol.https, description="""protocol to use, defaults to https"""
    )


def build_domain(config: WebfingerLookupConfig, acct_uri: str) -> str:
    """Builds the domain to use for webfinger lookups

    ```pycon
    >>> build_domain(WebfingerLookupConfig(protocol=Protocol.http),
    ...     "acct:alice@server.example")
    'http://server.example'

    ```
    """
    return f"{config.protocol}://{acct_uri.split('@')[1]}"


extension = Extension("webfinger", module=__name__, config_class=WebfingerLookupConfig)


@extension.lookup()
async def lookup(
    lookup: Lookup,
    session: ClientSession,
    config: extension.Config,  # type:ignore
) -> Lookup:
    if lookup.uri.startswith("acct:"):
        uri, _ = await lookup_uri_with_webfinger(
            session, lookup.uri, domain=build_domain(config, lookup.uri)
        )

        if uri:
            lookup.uri = uri
    return lookup



from dataclasses import dataclass, field
from pydantic import BaseModel, Field


class RewriteRules(BaseModel):
    """Describes the rewrite rules"""

    rules: dict[str, str] = Field(description="Mapping from old rule to new rule")


@dataclass
class RewriteConfiguration:
    """Configuration of the rewrite process"""

    ruleset: dict[str, RewriteRules] = field(
        metadata={"description": "mapping between group names and rewrite rules"}
    )

    def rewrite(self, method_name: str, group_names: list[str]) -> str:
        """rewrites the method_name"""
        for group_name in group_names:
            rules = self.ruleset.get(group_name)

            if rules and method_name in rules.rules:
                return rules.rules[method_name]

        return method_name

    def add_rules(
        self, group_name: str, rules: dict[str, str], overwrite: bool = False
    ):
        """Adds new rules to the ruleset"""
        if not overwrite and group_name in self.ruleset:
            return

        self.ruleset[group_name] = RewriteRules.model_validate({"rules": rules})

    @staticmethod
    def from_rules(rules: dict[str, dict[str, str]] | None):
        """
        Constructs the RewriteConfiguration

        ```
        >>> rules = {"old": "new"}
        >>> config = RewriteConfiguration.from_rules({"group": rules})
        >>> config
        RewriteConfiguration(ruleset={'group': RewriteRules(rules={'old': 'new'})})

        >>> group_names = ["group"]
        >>> config.rewrite("unchanged", group_names)
        'unchanged'

        >>> config.rewrite("old", group_names)
        'new'

        ```
        """
        if rules:
            mapped = {
                group: RewriteRules.model_validate({"rules": rule})
                for group, rule in rules.items()
            }
        else:
            mapped = {}

        return RewriteConfiguration(ruleset=mapped)



import pytest

import dynaconf
from dynaconf import Dynaconf

from .validators import all_validators


@pytest.fixture(scope="module")
def settings():
    return Dynaconf(validators=all_validators)


@pytest.mark.parametrize(
    ["key", "value"],
    [
        ("internal_exchange", "cattle_grid_internal"),
        ("exchange", "cattle_grid"),
        ("account_exchange", "amq.topic"),
    ],
)
def test_activity_pub_validators(settings, key, value):
    assert settings.activity_pub[key] == value


@pytest.mark.parametrize(
    ["key", "value"],
    [("enable", False), ("enable_reset", False)],
)
def test_gateway_admin(settings, key, value):
    assert settings.gateway.admin[key] == value


@pytest.mark.parametrize(
    ["key", "value"],
    [
        ("forbidden_names", ["bovine", "cattle_grid", "admin", "guest"]),
        ("allowed_name_regex", "^[a-zA-Z0-9_]{1,16}$"),
    ],
)
def test_account(settings, key, value):
    assert settings.account[key] == value


@pytest.mark.parametrize(
    ["key", "value"],
    [
        ("require_signature_for_activity_pub", True),
    ],
)
def test_auth_config(settings, key, value):
    assert settings.auth[key] == value


@pytest.mark.parametrize(
    ["key", "value"],
    [("enable", False), ("accounts", [])],
)
def test_testing_config(settings, key, value):
    assert settings.testing[key] == value


@pytest.mark.parametrize(
    ["key", "value"],
    [
        ("amqp_uri", "amqp://localhost"),
        ("db_uri", "sqlite:///cattle_grid.db"),
        ("enable_reporting", False),
        ("processor_in_app", False),
        ("permissions", {}),
    ],
)
def test_base_validators(settings, key, value):
    assert settings[key] == value


def test_plugins(settings):
    assert settings.plugins == []


@pytest.mark.parametrize(
    ["key", "value"],
    [
        ("base_urls", []),
    ],
)
def test_frontend_validators(settings, key, value):
    assert settings.frontend[key] == value


def test_frontend_validations(settings):
    settings.update({"frontend.base_urls": ["http://abel"]}, validate=True)
    settings.update({"frontend.base_urls": ["https://abel"]}, validate=True)

    with pytest.raises(dynaconf.validator.ValidationError):
        settings.update({"frontend.base_urls": ["abel"]}, validate=True)



import pytest
import click

from unittest.mock import MagicMock

from click.testing import CliRunner

from cattle_grid.testing.fixtures import *  # noqa
from cattle_grid.testing.cli import *  # noqa

from .account import account_with_name_password, list_permissions
from .cli import add_account_commands, new_account, modify_permissions


@pytest.fixture
def cli(db_uri):
    config = MagicMock()

    config.db_uri = db_uri

    @click.group()
    @click.pass_context
    def main(ctx):
        ctx.ensure_object(dict)
        ctx.obj["config"] = config

    add_account_commands(main)

    return main


def test_new_account_cli(create_database, cli, tmp_path):
    runner = CliRunner()
    result = runner.invoke(cli, ["account", "new", "user", "pass"])

    assert result.exit_code == 0


async def test_new_account(sql_session_for_test):
    await new_account(sql_session_for_test, "user", "pass", permission=["one"])

    account = await account_with_name_password(sql_session_for_test, "user", "pass")

    assert account
    await sql_session_for_test.refresh(account, attribute_names=["permissions"])
    permissions = list_permissions(account)

    assert permissions == ["one"]


async def test_modify_permissions(sql_session_for_test):
    await new_account(sql_session_for_test, "user", "pass", permission=["one"])
    await modify_permissions(sql_session_for_test, "user", ["two"], ["one"])

    account = await account_with_name_password(sql_session_for_test, "user", "pass")

    assert account
    await sql_session_for_test.refresh(account, attribute_names=["permissions"])
    permissions = list_permissions(account)

    assert permissions == ["two"]



import os

import asyncio
import logging

from behave import given, when, then

from almabtrieb import Almabtrieb

from cattle_grid.config import load_settings
from cattle_grid.account.account import create_account, add_permission
from cattle_grid.testing.features import publish_as
from cattle_grid.database import database_session

logger = logging.getLogger(__name__)


@given('An account called "{alice}"')  # type: ignore
async def create_account_step(context, alice: str):
    """
    Creates an account with name `alice`

    ```gherkin
    Given An account called "Alice"
    ```

    The connection is stored in `context.connections[alice]` it
    is an [Almabtrieb][almabtrieb.Almabtrieb] object.
    """

    config = load_settings()
    async with database_session(db_uri=config.get("db_uri")) as session:  # type: ignore
        account = await create_account(session, alice, alice)
        assert account
        await add_permission(session, account, "admin")

    mq_host = os.environ.get("CATTLE_GRID_MQ", "rabbitmq")

    context.connections[alice] = Almabtrieb.from_connection_string(
        f"amqp://{alice}:{alice}@{mq_host}/", silent=True
    )
    context.connections[alice].task = asyncio.create_task(
        context.connections[alice].run()
    )

    for _ in range(10):
        if context.connections[alice].connected:
            return
        logger.debug("Connecting")
        await asyncio.sleep(0.1)

    raise RuntimeError("Could not connect")


@given('"{alice}" created an actor called "{alyssa}"')  # pyright: ignore[reportCallIssue]
def create_actor_step(context, alice, alyssa):
    """
    ```gherkin
    Given "Alice" created an actor called "Alyssa"
    ```
    """
    hostname = {"alice": "abel", "bob": "banach", "Bob": "banach"}.get(alyssa, "abel")

    context.execute_steps(
        f"""
        Given "{alice}" created an actor on "{hostname}" called "{alyssa}"
        """
    )


@given('"{alice}" created an actor on "{hostname}" called "{alyssa}"')  # pyright: ignore[reportCallIssue]
async def create_actor_on_host_step(context, alice, alyssa, hostname):
    result = await context.connections[alice].create_actor(
        base_url=f"http://{hostname}", preferred_username=alyssa
    )
    context.actors[alyssa] = result


@when('"{alice}" deletes the actor "{alyssa}"')  # pyright: ignore[reportCallIssue]
async def alice_delete_actor_step(context, alice, alyssa):
    alyssa_id = context.actors[alyssa].get("id")

    await publish_as(
        context,
        alice,
        "delete_actor",
        {
            "actor": alyssa_id,
        },
    )


@then('"{alice}" has no actors')  # pyright: ignore[reportCallIssue]
async def check_no_actors(context, alice):
    response = await context.connections[alice].info()

    assert len(response.actors) == 0, f"Alice has actors {', '.join(response.actors)}"



from fastapi import FastAPI

from faststream import FastStream
from faststream.rabbit import RabbitBroker
from faststream.asyncapi import get_app_schema
from faststream.asyncapi.schema import Schema

from . import Extension
from .load.lifespan import lifespan_for_extension


def async_schema_for_extension(extension: Extension) -> Schema:
    """Converts the activity router of the extension to
    a [Schema][faststream.asyncapi.schema.Schema].

    this can be converted to a json string via
    `result.to_json()`."""
    broker = RabbitBroker()

    if extension.activity_router:
        broker.include_router(extension.activity_router)

    app = FastStream(broker, title=extension.name)

    return get_app_schema(app)


def fastapi_for_extension(
    extension: Extension, include_broker: bool = False
) -> FastAPI:
    """Converts the api router of the extension to a
    [FastAPI][fastapi.FastAPI] app."""

    app = FastAPI(
        title=extension.name,
        description=extension.description or "",
        lifespan=lifespan_for_extension(extension, include_broker=include_broker),
    )

    app.include_router(extension.api_router)
    return app


def openapi_schema_for_extension(extension: Extension) -> dict:
    """Converts the api router of the extension to a
    dictionary containing the openapi schema"""
    app = fastapi_for_extension(extension)

    return app.openapi()



from unittest.mock import MagicMock
from contextlib import asynccontextmanager

from .lifespan import iterate_lifespans


def call_from_context(func):
    @asynccontextmanager
    async def context():
        func()
        yield

    return context


async def test_iterate_lifespans_empty_list():
    async with iterate_lifespans([]):
        pass


async def test_iterate_lifespans_one_item():
    one = MagicMock()
    async with iterate_lifespans([call_from_context(one)]):  # type: ignore
        pass

    one.assert_called_once()


async def test_iterate_lifespans_three_items():
    mocks = [MagicMock(), MagicMock(), MagicMock()]

    async with iterate_lifespans([call_from_context(x) for x in mocks]):  # type: ignore
        pass

    for mock in mocks:
        mock.assert_called_once()



from typing import Annotated

from fast_depends import Depends
from sqlalchemy import select
from sqlalchemy.orm import joinedload
from sqlalchemy.ext.asyncio import AsyncSession

from cattle_grid.database.account import Account, ActorForAccount
from cattle_grid.database.activity_pub_actor import Actor
from cattle_grid.model.account import WithActor
from cattle_grid.dependencies import SqlSession
from cattle_grid.dependencies.processing import RoutingKey


def name_from_routing_key(
    routing_key: RoutingKey,
) -> str:
    """
    ```pycon
    >>> name_from_routing_key("receiving.alice")
    'alice'

    >>> name_from_routing_key("receiving.alice.action.fetch")
    'alice'

    ```
    """
    return routing_key.split(".")[1]


AccountName = Annotated[str, Depends(name_from_routing_key)]
"""Assigns the account name extracted from the routing key"""


def method_from_routing_key(
    name: AccountName,
    routing_key: RoutingKey,
) -> str:
    """
    Extracts the method from the routing key

    ```pycon
    >>> method_from_routing_key("alice", "send.alice.trigger.method.first")
    'method.first'

    ```
    """
    start_string = f"send.{name}.trigger."
    if routing_key.startswith(start_string):
        return routing_key.removeprefix(start_string)
    else:
        raise ValueError("Invalid routing key for trigger")


MethodFromRoutingKey = Annotated[str, Depends(method_from_routing_key)]
"""Returns the method of a trigger message"""


async def account(name: AccountName, session: SqlSession) -> Account:
    account = await session.scalar(
        select(Account).where(Account.name == name).options(joinedload(Account.actors))
    )
    if account is None:
        raise ValueError("Account not found for name %s", name)

    return account


AccountFromRoutingKey = Annotated[Account, Depends(account)]
"""Returns the account from the routing key"""


async def actor_for_account_from_account(
    msg: WithActor, account: Annotated[Account, Depends(account)]
) -> ActorForAccount | None:
    for actor in account.actors:
        if actor.actor == msg.actor:
            return actor
    return None


ActorForAccountFromMessage = Annotated[
    ActorForAccount, Depends(actor_for_account_from_account)
]
"""The actor provided in the send message"""


async def actor_from_account(
    account: Account, actor_id: str, session: AsyncSession
) -> Actor | None:
    for actor in account.actors:
        if actor.actor == actor_id:
            return await session.scalar(select(Actor).where(Actor.actor_id == actor_id))
    return None


async def actor(
    msg: WithActor, session: SqlSession, account: Account = Depends(account)
):
    actor = await actor_from_account(account, msg.actor, session)
    if not actor:
        raise ValueError(
            f"Actor not found for account name {account.name} and actor id {msg.actor}"
        )
    return actor


ActorFromMessage = Annotated[Actor, Depends(actor)]
"""The actor provided in the send message"""



import logging


def configure_logging(settings):
    """Sets the log level according to the settings"""
    log_config = settings.get("logging", {})

    for name, level in log_config.items():
        logging.getLogger(name).setLevel(level.upper())



import asyncio
import logging
import aio_pika

from typing import AsyncGenerator
from contextlib import asynccontextmanager
from uuid import uuid4

logger = logging.getLogger(__name__)


@asynccontextmanager
async def queue_for_connection(
    connection: aio_pika.RobustConnection,
    routing_key: str,
    name: str,
    exchange_name: str,
) -> AsyncGenerator[aio_pika.abc.AbstractQueue, None]:
    async with connection:
        channel = await connection.channel()
        await channel.set_qos(prefetch_count=1)

        exchange = await channel.get_exchange(exchange_name)

        queue = await channel.declare_queue(
            f"queue-{name}-{uuid4()}",
            durable=False,
            auto_delete=True,
            exclusive=True,
        )
        await queue.bind(exchange, routing_key=routing_key)

        yield queue


async def enqueue_to_routing_key_and_connection(
    connection: aio_pika.RobustConnection,
    asyncio_queue: asyncio.Queue,
    routing_key: str,
    name: str = "cattle-grid",
    exchange_name: str = "amq.topic",
):
    """Subscribes to the routing key and adds all
    the received messages to `asyncio_queue` as a string"""

    async with queue_for_connection(
        connection, routing_key=routing_key, name=name, exchange_name=exchange_name
    ) as queue:
        async with queue.iterator() as iterator:
            try:
                async for message in iterator:
                    async with message.process():
                        message_body = message.body.decode()
                        logger.info(message_body)
                        await asyncio_queue.put(message_body)
            except Exception as e:
                logger.exception(e)



from unittest.mock import MagicMock
from fast_depends import inject

from . import Config


def test_config():
    mock = MagicMock()

    def tester(config: Config):
        mock(config.db_uri)

    inject(tester)()  # type: ignore

    mock.assert_called_once()



from faststream import FastStream
from faststream.rabbit import RabbitBroker

import logging

from contextlib import asynccontextmanager

from .app import add_routers_to_broker, init_extensions

from .dependencies.globals import global_container

from .extensions.load import (
    lifespan_from_extensions,
)
from .exchange.exception import exception_middleware

logging.basicConfig(level=logging.DEBUG)

global_container.load_config()
extensions = init_extensions(global_container.config)

broker = RabbitBroker(
    global_container.config.amqp_uri,  # type:ignore
    middlewares=[exception_middleware],
)
add_routers_to_broker(broker, extensions, global_container.config)


@asynccontextmanager
async def lifespan():
    async with global_container.common_lifecycle(global_container.config):
        async with lifespan_from_extensions(extensions):
            yield


app = FastStream(broker, lifespan=lifespan)


@app.after_startup
async def declare_exchanges() -> None:
    await broker.declare_exchange(global_container.internal_exchange)
    await broker.declare_exchange(global_container.exchange)



from unittest.mock import AsyncMock
import aiohttp

from cattle_grid.dependencies.globals import global_container
from cattle_grid.model.lookup import Lookup

from . import build_lookup, load_extension, collect_method_information


async def test_build_lookup():
    extensions = [
        load_extension({"module": "cattle_grid.extensions.examples.webfinger_lookup"})
    ]

    lookup = build_lookup(extensions)

    data = Lookup(uri="http://remote.example", actor="http://actor.example")

    global_container.session = AsyncMock(aiohttp.ClientSession)

    await lookup(data)

    global_container.session = None


async def test_collect_method_information():
    extensions = [
        load_extension({"module": "cattle_grid.extensions.examples.simple_storage"})
    ]

    method_information = collect_method_information(extensions)

    assert len(method_information) == 2

    assert {x.routing_key for x in method_information} == {
        "publish_activity",
        "publish_object",
    }



"""
This cache stores objects temporally in a key value database

Sample configuration

```toml title="cattle_grid.toml"
[[extensions]]
module = "cattle_grid.extensions.examples.cache"
lookup_order = 4
config = { connection_url = "redis://cattle_grid_redis", duration = 3600 }
```

The duration is expressed in seconds. By using the special
connection_url `:memory:` the cache will be replaced by
[fakeredis](https://fakeredis.readthedocs.io/en/latest/).
This means that the cache is in memory and no external key value
database is necessary. This is useful for testing, but not
recommended for production use.

"""

import logging
import json

from cattle_grid.model import ActivityMessage
from .dependencies import CacheClient

from contextlib import asynccontextmanager
from cattle_grid.extensions import Extension

from redis import asyncio as redis

from .lookup import cache_lookup
from .config import CacheConfiguration

logger = logging.getLogger(__name__)


extension = Extension(
    name="public data cache",
    module=__name__,
    config_class=CacheConfiguration,
)


@asynccontextmanager
async def lifespan(config: extension.Config):  # type: ignore
    import cattle_grid.extensions.examples.cache.dependencies as dependencies

    if config.connection_url == ":memory:":
        logger.info("Using in memory cache")

        import fakeredis

        async with fakeredis.FakeAsyncRedis() as client:
            dependencies.cache = client
            yield
    else:
        client = redis.from_url(config.connection_url)

        logger.info("Connected to redis %s", config.connection_url)

        dependencies.cache = client

        yield


extension.lifespan = lifespan


@extension.subscribe("incoming.Create")
async def handle_create_activity(
    activity: ActivityMessage,
    cache: CacheClient,
    config: extension.Config,  # type: ignore
) -> None:
    obj = activity.data.get("raw", {}).get("object")

    await cache.setex(obj.get("id"), config.duration, json.dumps(obj))


extension.lookup()(cache_lookup)



from enum import StrEnum, auto
from pydantic import BaseModel, ConfigDict, Field


class UpdateActionType(StrEnum):
    """Available actions for updating the actor"""

    add_identifier = auto()
    """Adds a new identifier. The identifier is assumed to already exist."""
    create_identifier = auto()
    """Creates a new identifier. Must be on a domain controlled by cattle_grid and enabled in the account"""
    update_identifier = auto()
    """Updates an identifer"""
    remove_identifier = auto()
    """Removes an identifier"""

    rename = auto()
    """Updates the internal name of the actor"""

    update_property_value = auto()
    """Adds or updates a property value of the actor"""
    remove_property_value = auto()
    """Removes a property value"""

    add_url = auto()
    """Adds an url to the actor profile"""
    remove_url = auto()
    """Removes an url from the actor profile"""


class UpdateAction(BaseModel):
    """Action to update an actor"""

    model_config = ConfigDict(
        extra="allow",
    )

    action: UpdateActionType


class UpdateIdentifierAction(UpdateAction):
    """Used to update an identifier of the actor"""

    identifier: str = Field(
        description="The identifier", examples=["acct:alice@domain.example"]
    )
    primary: bool = Field(
        False,
        description="Set the identifier as the primary one, if the identifier corresponds to an acct-uri this will update the primary identifier",
    )


class RenameActorAction(UpdateAction):
    """Update the internal name of the actor"""

    name: str = Field(description="The new name of the actor")


class UpdatePropertyValueAction(UpdateAction):
    """Update a property value of the actor"""

    key: str = Field(
        examples=["author"],
        description="The key of the property value to be created, updated, or deleted",
    )
    value: str | None = Field(
        None,
        examples=["Alice"],
        description="The value of the property value",
    )


class UpdateUrlAction(UpdateAction):
    url: str = Field(description="The url to add")
    media_type: str | None = Field(
        default=None, description="The media type for the url"
    )
    rel: str | None = Field(default=None, description="The relation to use")



from pydantic import BaseModel, Field


class WithActor(BaseModel):
    actor: str = Field(
        description="actor_id of the actor that received the message",
        examples=["http://host.example/actor"],
    )



from contextlib import asynccontextmanager
from typing import Callable
from faststream.rabbit import RabbitBroker, TestRabbitBroker, RabbitQueue

from cattle_grid.dependencies.globals import global_container


from .load import add_routers_to_broker


@asynccontextmanager
async def with_test_broker_for_extension(
    extensions, subscribers: dict[str, Callable] = {}
):
    """Creates a test broker with subscribtions to the routing_keys given in subscriber.

    ```python
    my_mock = AsyncMock()

    async with with_test_broker_for_extension([extension], {
        "routing_key_to_listen_to": my_mock
    })
    ```
    """
    broker = RabbitBroker()

    for routing_key, subscriber in subscribers.items():
        broker.subscriber(
            RabbitQueue(routing_key, routing_key=routing_key),
            exchange=global_container.exchange,
        )(subscriber)

    add_routers_to_broker(broker, extensions)
    async with TestRabbitBroker(broker) as tbr:
        yield tbr



from faststream import Context
from cattle_grid.database.activity_pub_actor import StoredActivity

from uuid6 import uuid7
from datetime import datetime

from cattle_grid.model.processing import StoreActivityMessage
from cattle_grid.dependencies.globals import global_container
from cattle_grid.dependencies.processing import MessageActor
from cattle_grid.dependencies import CommittingSession


async def store_activity_subscriber(
    message: StoreActivityMessage,
    actor: MessageActor,
    session: CommittingSession,
    broker=Context(),
):
    """This method used internally to store activities
    generated by `cattle_grid`, e.g. an automatically generated
    accept requests. We note that it assigns an id to
    the activity.

    After storing the activity, it is published to the
    appropriate `outgoing.*` topic, which triggers sending
    the activity.
    """
    base_url = "/".join(actor.actor_id.split("/")[:-2])
    uuid = str(uuid7())
    activity_id = f"{base_url}/object/{uuid}"

    activity = message.data
    activity["id"] = activity_id
    activity_type = activity.get("type")

    session.add(
        StoredActivity(id=uuid, actor=actor, data=activity, published=datetime.now())
    )

    await broker.publish(
        {"actor": actor.actor_id, "data": activity},
        routing_key=f"outgoing.{activity_type}",
        exchange=global_container.internal_exchange,
    )



"""
This extension is an example of storing activities
and then serving them through a HTTP API.

I will possibly extend it to also store objects
(and provide some `Create` activity creation), but
not much more.

A real storage mechanism should have several features
this simple API has not, e.g.

- Serving a HTML page through content negotiation
- Allowing one to update the content of the database
- Collecting and adding metadata, e.g. a replies collection for objects

Usage:

```toml
[[extensions]]
module = "cattle_grid.extensions.examples.simple_storage"
api_prefix = "/simple_storage"

config = { prefix = "/simple_storage/" }
```
"""

import logging
import uuid

from fastapi import HTTPException
from faststream import Context
from faststream.rabbit import RabbitBroker
from sqlalchemy.future import select

from cattle_grid.dependencies.globals import global_container
from cattle_grid.dependencies import CorrelationId, CommittingSession
from cattle_grid.dependencies.fastapi import SqlSession as FastApiSession
from cattle_grid.dependencies.processing import FactoriesForActor

from cattle_grid.extensions import Extension
from cattle_grid.extensions.util import lifespan_for_sql_alchemy_base_class
from cattle_grid.tools.fastapi import ActivityResponse
from cattle_grid.model import ActivityMessage

from cattle_grid.activity_pub.actor.requester import (
    ActorNotFound,
    is_valid_requester_for_obj,
)
from cattle_grid.activity_pub.server.router import ActivityPubHeaders

from .models import StoredActivity, StoredObject, Base
from .message_types import PublishActivity, PublishObject
from .config import SimpleStorageConfiguration

logger = logging.getLogger(__name__)

extension = Extension(
    name="simple storage",
    module=__name__,
    lifespan=lifespan_for_sql_alchemy_base_class(Base),
    config_class=SimpleStorageConfiguration,
)
"""The simple storage extension"""


@extension.subscribe("publish_activity")
async def simple_storage_publish_activity(
    message: PublishActivity,
    config: extension.Config,  # type: ignore
    session: CommittingSession,
    correlation_id: CorrelationId,
    broker: RabbitBroker = Context(),
):
    """Method to subscribe to the `publish_activity` routing_key.

    An activity send to this endpoint will be stored in the
    database, and then published through the `send_message`
    mechanism.

    The stored activity can then be retrieved through the
    HTTP API.
    """
    if message.data.get("id"):
        raise ValueError("Activity ID must not be set")

    if message.data.get("actor") != message.actor:
        raise ValueError("Actor must match message actor")

    activity = message.data
    activity["id"], uuid = config.make_id(message.actor)

    logger.info("Publishing activity with id %s for %s", message.actor, activity["id"])

    session.add(
        StoredActivity(
            id=uuid,
            data=activity,
            actor=message.actor,
        )
    )

    await broker.publish(
        ActivityMessage(actor=message.actor, data=activity),
        routing_key="send_message",
        exchange=global_container.exchange,
        correlation_id=correlation_id,
    )


@extension.subscribe("publish_object")
async def simple_storage_publish_object(
    message: PublishObject,
    config: extension.Config,  # type: ignore
    session: CommittingSession,
    factories: FactoriesForActor,
    correlation_id: CorrelationId,
    broker: RabbitBroker = Context(),
):
    """Publishes an object, subscribed to the routing key
    `publish_object`.

    We note that this routine creates a `Create` activity for the object."""

    obj = message.data

    if obj.get("id"):
        raise ValueError("Object ID must not be set")

    if obj.get("attributedTo") != message.actor:
        raise ValueError("Actor must match object attributedTo")

    obj["id"], obj_uuid = config.make_id(message.actor)

    logger.info("Publishing object with id %s for %s", message.actor, obj["id"])

    activity = factories[0].create(obj).build()

    await broker.publish(
        ActivityMessage(actor=message.actor, data=activity),
        routing_key="publish_activity",
        exchange=global_container.exchange,
        correlation_id=correlation_id,
    )
    session.add(
        StoredObject(
            id=obj_uuid,
            data=obj,
            actor=message.actor,
        )
    )


@extension.get("/")
async def main():
    """Basic endpoint that just returns a string, so
    requesting with an uuid doesn't return an error."""
    return "simple storage cattle grid sample extension"


@extension.get("/{uuid}", response_class=ActivityResponse)
async def get_activity_or_object(
    uuid: uuid.UUID, headers: ActivityPubHeaders, session: FastApiSession
):
    """Returns the activity or object"""
    result = await session.scalar(
        select(StoredActivity).where(StoredActivity.id == uuid)
    )

    if result is None:
        result = await session.scalar(
            select(StoredObject).where(StoredObject.id == uuid)
        )

    if result is None:
        raise HTTPException(status_code=404, detail="Activity not found")

    try:
        if not headers.x_cattle_grid_requester or not await is_valid_requester_for_obj(
            session, headers.x_cattle_grid_requester, result.data
        ):
            raise HTTPException(status_code=401)
    except ActorNotFound:
        raise HTTPException(status_code=410, detail="Activity no longer available")

    if result.data.get("id") != headers.x_ap_location:
        raise HTTPException(status_code=400, detail="Location header does not match id")

    return result.data



from sqlalchemy import String, Text
from sqlalchemy.ext.asyncio import AsyncAttrs
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column


class Base(AsyncAttrs, DeclarativeBase):
    """Base model"""

    pass


class Credential(Base):
    """Stored credential"""

    __tablename__ = "credential"
    """name of the table"""

    id: Mapped[int] = mapped_column(primary_key=True)
    """primary key"""

    actor_id: Mapped[str] = mapped_column(String(256))
    """The id of the actor the key belongs to"""
    identifier: Mapped[str] = mapped_column(String(256), unique=True)
    """The identifier of the key"""

    secret: Mapped[str] = mapped_column(Text())
    """The secret underlying the private key"""


class InboxLocation(Base):
    """Describes the location of an inbox. Used to send
    ActivityPub Activities addressed to the actor to the
    corresponding inbox.

    This information is also collected for remote actors.
    """

    __tablename__ = "inboxlocation"

    id: Mapped[int] = mapped_column(primary_key=True)
    actor: Mapped[str] = mapped_column(String(256), unique=True)
    inbox: Mapped[str] = mapped_column(String(256))



"""
Extension to add information to the context

See [FEP-f228](https://codeberg.org/fediverse/fep/src/branch/main/fep/f228/fep-f228.md)

This extension requires the muck out extension to be installed.
"""

import logging

from sqlalchemy import select
from uuid6 import uuid7
from cattle_grid.dependencies import CommittingSession
from cattle_grid.extensions import Extension

from muck_out.types.validated import Object


from cattle_grid.extensions.util import lifespan_for_sql_alchemy_base_class
from .models import Base, ContextInformation


logger = logging.getLogger(__name__)

extension = Extension(
    "context processor", __name__, lifespan=lifespan_for_sql_alchemy_base_class(Base)
)


@extension.transform(inputs=["parsed"], outputs=["context"])
async def transform(
    data: dict[str, dict], session: CommittingSession
) -> dict[str, dict]:
    parsed = data["parsed"]
    activity_type = parsed.get("activity", {}).get("type")

    if activity_type != "Create":
        return {"context": {}}

    try:
        obj = Object.model_validate(parsed.get("embeddedObject"))
    except Exception as e:
        logger.exception(e)
        return {"context": {}}

    in_context = await session.scalar(
        select(ContextInformation).where(ContextInformation.object_id == obj.id)
    )
    if in_context:
        return {"context": {"id": str(in_context.context_id)}}

    reply_for = await session.scalar(
        select(ContextInformation).where(
            ContextInformation.object_id == obj.in_reply_to
        )
    )

    if reply_for:
        context_id = reply_for.context_id
        session.add(
            ContextInformation(
                context_id=context_id, object_id=obj.id, parent_id=reply_for.object_id
            )
        )
    else:
        context_id = uuid7()
        session.add(ContextInformation(context_id=context_id, object_id=obj.id))

    return {"context": {"id": str(context_id)}}



import pytest

from unittest.mock import AsyncMock

from faststream.rabbit import RabbitBroker, TestRabbitBroker, RabbitQueue

from cattle_grid.testing.fixtures import *  # noqa
from cattle_grid.dependencies.globals import global_container

from .router import create_router
from cattle_grid.database.account import Account


@pytest.fixture
async def subscriber_mock():
    return AsyncMock()


@pytest.fixture
async def test_account():
    return await Account.create(name="alice", password_hash="password")


@pytest.fixture
async def test_broker(subscriber_mock):
    br = RabbitBroker("amqp://guest:guest@localhost:5672/")
    br.include_router(create_router())

    br.subscriber(
        RabbitQueue("error-queue", routing_key="error.#"),
        exchange=global_container.account_exchange,
    )(subscriber_mock)

    async with TestRabbitBroker(br) as tbr:
        yield tbr


async def test_exception(test_broker, subscriber_mock, actor_with_account):
    await test_broker.publish(
        {},
        routing_key="send.alice.request.fetch",
        exchange=global_container.account_exchange,
    )

    subscriber_mock.assert_awaited_once()



import asyncio
import pytest
from sqlalchemy import select

from cattle_grid.testing.fixtures import *  # noqa
from cattle_grid.database.account import ActorForAccount

from .testing import *  # noqa
from .streaming import get_message_streamer


async def test_stream(test_app, test_client, bearer_header):
    def streamer(account_name, event_type):
        queue = asyncio.Queue()

        async def putter():
            await queue.put("hello")
            await queue.put(account_name)
            await queue.put(None)

        return queue, asyncio.create_task(putter())

    test_app.dependency_overrides[get_message_streamer] = lambda: streamer

    response = test_client.get("/account/stream/incoming", headers=bearer_header)
    assert response.status_code == 200

    assert response.headers["content-type"].split(";")[0] == "text/event-stream"

    assert (
        response.text
        == """data: hello

data: name

"""
    )


@pytest.mark.parametrize("name,expected", [(None, "from_api"), ("alice", "alice")])
async def test_create_actor(
    sql_session_for_test, test_client, bearer_header, name, expected
):
    result = test_client.post(
        "/account/create",
        json={"baseUrl": "http://abel.test", "name": name},
        headers=bearer_header,
    )
    assert result.status_code == 201

    actor_id = result.json()["id"]

    actor = await sql_session_for_test.scalar(
        select(ActorForAccount).where(ActorForAccount.actor == actor_id)
    )
    assert actor
    assert actor.name == expected


def test_create_actor_with_handle(test_client, bearer_header):
    result = test_client.post(
        "/account/create",
        json={"baseUrl": "http://abel.test", "handle": "alice"},
        headers=bearer_header,
    )
    assert result.status_code == 201

    actor = result.json()
    assert actor["preferredUsername"] == "alice"


def test_create_actor_with_handle_duplicate(test_client, bearer_header):
    result = test_client.post(
        "/account/create",
        json={"baseUrl": "http://abel.test", "handle": "alice"},
        headers=bearer_header,
    )
    assert result.status_code == 201

    result = test_client.post(
        "/account/create",
        json={"baseUrl": "http://abel.test", "handle": "alice"},
        headers=bearer_header,
    )
    assert result.status_code == 409



from typing import Callable, Awaitable, Dict, Set
from dataclasses import dataclass


@dataclass
class Transformer:
    """Data model for a transformer"""

    name: str
    """name of the transformer"""

    transformer: Callable[[Dict], Awaitable[Dict]]
    """method that transforms the data"""

    inputs: Set[str]
    """list of required input fields"""

    outputs: Set[str]
    """list of output fields"""

    def __hash__(self):
        return hash(self.name)



import tomli_w
from . import load_settings


def test_load_config_new_file(tmp_path):
    config = load_settings(tmp_path / "doesnotexist.toml")

    assert not config.enable_reporting


def test_load_config_existing_file(tmp_path):
    filename = tmp_path / "config.toml"
    with open(filename, "wb") as fp:
        tomli_w.dump({"enable_reporting": True}, fp)
    config = load_settings(filename)

    assert config.enable_reporting



import pytest
import asyncio

from cattle_grid.database import database_engine


@pytest.fixture
def db_uri(tmp_path):
    return "sqlite+aiosqlite:///" + str(tmp_path / "test.db")


@pytest.fixture
def create_database(db_uri):
    async def run_method():
        async with database_engine(db_uri=db_uri) as engine:
            from cattle_grid.database.activity_pub import Base as APBase

            async with engine.begin() as conn:
                await conn.run_sync(APBase.metadata.create_all)

        await asyncio.sleep(0.3)

    asyncio.run(run_method())



import uuid
import logging

from typing import Annotated
from fastapi import Depends, HTTPException
from sqlalchemy import select
from sqlalchemy.orm import joinedload

from cattle_grid.activity_pub.actor import actor_to_object
from cattle_grid.database.activity_pub_actor import Actor
from cattle_grid.dependencies.fastapi import SqlSession
from .models import PublishedObject, PublishingActor

logger = logging.getLogger(__name__)


async def publishing_actor_for_name(session: SqlSession, actor_name: str):
    actor = await session.scalar(
        select(PublishingActor).where(PublishingActor.name == actor_name)
    )

    if not actor:
        raise HTTPException(404)

    return actor


PublishingActorForName = Annotated[PublishingActor, Depends(publishing_actor_for_name)]
"""Returns the publishing actor"""


async def published_object(session: SqlSession, uuid: uuid.UUID):
    obj = await session.scalar(
        select(PublishedObject).where(PublishedObject.id == uuid)
    )
    if not obj:
        raise HTTPException(404)

    return obj


PublishedObjectForUUID = Annotated[PublishedObject, Depends(published_object)]
"""Returns the published object"""


async def get_actor_profile(
    session: SqlSession, publishing_actor: PublishingActorForName
):
    actor = await session.scalar(
        select(Actor)
        .where(Actor.actor_id == publishing_actor.actor)
        .options(joinedload(Actor.identifiers))
    )

    if actor is None:
        logger.warning(
            "Could not find actor for publishing actor with name %s",
            publishing_actor.name,
        )
        raise HTTPException(404)

    return actor_to_object(actor)


ActorProfile = Annotated[dict, Depends(get_actor_profile)]
"""Returns the actor profile"""



import jsonschema
import json


from behave import then, when, given

from urllib.parse import urljoin

from bovine.models import JrdData


@given("A Fediverse server")
def fediverse_server(context):
    context.server = "http://abel"


@when('Querying "{path}"')
async def query_path(context, path):
    response = await context.session.get(urljoin(context.server, path))

    response.raise_for_status()
    context.response_body = await response.json()


@then('The jrd links contain one of type "{link_type}"')
def parse_jrd_links(context, link_type):
    data = JrdData.model_validate(context.response_body)

    value = None
    for link in data.links:
        if link.type == link_type:
            value = link.href

    assert value, f"Link not found in JRD {context.response_body}"

    context.link_to_resolve = value


@then("This page can be resolved")
async def can_resolve_page(context):
    response = await context.session.get(context.link_to_resolve)

    response.raise_for_status()
    context.result = await response.json()


@then('The result obeys the schema given by "{schema_url}"')
async def validate_against_schema(context, schema_url):
    response = await context.session.get(
        schema_url, headers={"accept": "application/json"}
    )
    schema = json.loads(await response.text())

    jsonschema.validate(context.result, schema)



import logging
import re

import aiohttp
from dataclasses import dataclass
from typing import Callable, Awaitable, Dict, List
from contextlib import asynccontextmanager
from functools import cached_property


from sqlalchemy.ext.asyncio import (
    AsyncSession,
    async_sessionmaker,
    create_async_engine,
    AsyncEngine,
)
from dynaconf import Dynaconf
from dynaconf.utils import DynaconfDict

from faststream.rabbit import (
    RabbitBroker,
    RabbitExchange,
    ExchangeType,
)

from cattle_grid.config.logging import configure_logging
from cattle_grid.config.rewrite import RewriteConfiguration
from cattle_grid.config import load_settings, default_filenames
from cattle_grid.model.lookup import LookupMethod
from cattle_grid.model.extension import MethodInformationModel

logger = logging.getLogger(__name__)


@dataclass
class GlobalContainer:
    session: aiohttp.ClientSession | None = None
    engine: AsyncEngine | None = None
    method_information: List[MethodInformationModel] | None = None

    transformer: Callable[[Dict], Awaitable[Dict]] | None = None
    lookup: LookupMethod | None = None

    _config: Dynaconf | DynaconfDict | None = None
    _rewrite_rules: RewriteConfiguration | None = None

    async_session_maker: Callable[[], AsyncSession] | None = None

    def __post_init__(self):
        self.load_config()

    def get_config(self):
        if self._config is None:
            raise ValueError("Config not loaded")
        return self._config

    @property
    def config(self):
        if self._config is None:
            raise ValueError("Config not loaded")
        return self._config

    def load_config(self, filenames: list[str] = default_filenames):
        self._config = load_settings(filenames)
        self._rewrite_rules = RewriteConfiguration.from_rules(
            self._config.get("rewrite")  # type: ignore
        )
        configure_logging(self._config)

    @asynccontextmanager
    async def session_lifecycle(self):
        async with aiohttp.ClientSession() as session:
            self.session = session
            yield session
            self.session = None

    @asynccontextmanager
    async def alchemy_database(self, db_uri: str | None = None, echo: bool = False):
        """Initializes the sql alchemy engine"""

        if db_uri is None:
            db_uri = self.config.db_uri  # type:ignore

        if db_uri is None:
            raise ValueError("Database URI not set")

        if self.engine or self.async_session_maker:
            raise ValueError("Database already initialized")

        self.engine = create_async_engine(db_uri, echo=echo)
        self.async_session_maker = async_sessionmaker(self.engine)
        logger.debug(
            "Connected to %s with sqlalchemy", re.sub("://.*@", "://***:***@", db_uri)
        )

        yield self.engine

        await self.engine.dispose()
        self.engine = None
        self.async_session_maker = None

    @asynccontextmanager
    async def common_lifecycle(self, config=None):
        if config is None:
            config = self.config

        async with self.session_lifecycle():
            async with self.alchemy_database():
                yield

    @cached_property
    def internal_exchange(self) -> RabbitExchange:
        """The internal exchange used to process
        ActivityPub messages related to the social graph

        :returns:
        """
        return RabbitExchange(
            self.config.activity_pub.internal_exchange,  # type: ignore
            type=ExchangeType.TOPIC,
        )

    def get_internal_exchange(self) -> RabbitExchange:
        return self.internal_exchange

    @cached_property
    def exchange(self) -> RabbitExchange:
        """Returns the public exchange used to process

        :returns:
        """
        return RabbitExchange(
            self.config.activity_pub.exchange,  # type: ignore
            type=ExchangeType.TOPIC,
        )

    def get_exchange(self) -> RabbitExchange:
        return self.exchange

    @cached_property
    def account_exchange(self) -> RabbitExchange:
        """Returns the exchange used to distribute messages between accounts

        :returns:
        """
        exchange_name = self.config.activity_pub.account_exchange  # type: ignore

        durable = True if exchange_name == "amq.topic" else False

        return RabbitExchange(exchange_name, type=ExchangeType.TOPIC, durable=durable)  # type: ignore

    def get_account_exchange(self) -> RabbitExchange:
        return self.account_exchange

    @cached_property
    def broker(self) -> RabbitBroker:
        return RabbitBroker(self.config.amqp_uri)  # type: ignore

    def get_broker(self) -> RabbitBroker:
        return self.broker

    def get_session_maker(self) -> Callable[[], AsyncSession]:
        if not self.async_session_maker:
            raise ValueError("Alchemy database not initialized")
        return self.async_session_maker

    def get_rewrite_rules(self) -> RewriteConfiguration:
        if not self._rewrite_rules:
            raise ValueError("Rules not loaded")
        return self._rewrite_rules


global_container = GlobalContainer()


def get_transformer() -> Callable[[Dict], Awaitable[Dict]]:
    global global_container

    if not global_container.transformer:
        raise ValueError("Transformer not initialized")

    return global_container.transformer


def get_lookup() -> LookupMethod:
    global global_container

    if not global_container.lookup:
        raise ValueError("Lookup not initialized")

    return global_container.lookup


def get_engine() -> AsyncEngine:
    global global_container

    if not global_container.engine:
        raise ValueError("Engine not initialized")

    return global_container.engine


def get_method_information() -> List[MethodInformationModel]:
    global global_container

    if global_container.method_information is None:
        raise ValueError("Method information not initialized")

    return global_container.method_information



from faststream import FastStream
from faststream.asyncapi.schema import Schema
from fastapi import FastAPI

from cattle_grid.version import __version__

from .processing import create_processing_router
from .server import router


def get_mock_faststream_app() -> FastStream:
    """Creates a mock faststream app for ActivityPub processing"""

    from faststream.rabbit import RabbitBroker

    broker = RabbitBroker()
    broker.include_router(create_processing_router())

    return FastStream(
        broker,
        title="cattle_grid ActivityPub processing",
        version=__version__,
        description="Illustrates how cattle grid processes ActivityPub",
    )


def get_async_api_schema() -> Schema:
    """Returns the async api schema for cattle_grid ActivityPub processing"""
    from faststream.asyncapi import get_app_schema

    app = get_mock_faststream_app()

    return get_app_schema(app)


def get_fastapi_app() -> FastAPI:
    """Returns the fast api app for ActivityPub processing"""

    app = FastAPI(title="cattle_grid ActivityPub routes", version=__version__)
    app.include_router(router)

    return app



import logging

from faststream import FastStream
from faststream.rabbit import RabbitRouter, RabbitExchange, RabbitQueue
from faststream.asyncapi.schema import Schema
from typing import Any, List, Tuple

from cattle_grid.version import __version__
from cattle_grid.dependencies.globals import global_container

from .exception import exception_middleware
from .info import add_method_information
from .handlers import update_actor, delete_actor_handler
from .message_handlers import send_message, fetch_object
from .shovel import incoming_shovel, outgoing_shovel

# from .data_types import ActivityMessage

logger = logging.getLogger(__name__)


def create_router(
    main_exchange: RabbitExchange | None = None, include_shovels=True
) -> RabbitRouter:
    """Creates a router to be used to manage users"""

    if main_exchange is None:
        main_exchange = global_container.exchange

    router = RabbitRouter(middlewares=[exception_middleware])

    if include_shovels:
        router.subscriber(
            RabbitQueue("incoming_shovel", routing_key="incoming.#"),
            global_container.internal_exchange,
        )(incoming_shovel)
        router.subscriber(
            RabbitQueue("outgoing_shovel", routing_key="outgoing.#"),
            global_container.internal_exchange,
        )(outgoing_shovel)

    routing_config: List[Tuple[str, Any]] = [
        ("update_actor", update_actor),
        ("delete_actor", delete_actor_handler),
        ("send_message", send_message),
        ("fetch_object", fetch_object),
        ("add_method_information", add_method_information),
    ]

    prefix = "cattle_grid_"

    for routing_key, coroutine in routing_config:
        router.subscriber(
            RabbitQueue(prefix + routing_key, routing_key=routing_key),
            exchange=main_exchange,
            title=routing_key,
        )(coroutine)

    return router


def get_mock_faststream_app() -> FastStream:
    from faststream.rabbit import RabbitBroker

    broker = RabbitBroker()
    broker.include_router(create_router(include_shovels=False))

    # @broker.subscriber("incoming.#", exchange="cattle_grid")
    # async def incoming_subscriber(message: ActivityMessage) -> None:
    #     """
    #     This is a sample implementation of a subscriber
    #     for incoming messages. cattle_grid publishes on
    #     the routing key `incoming.ACTIVITY_TYPE` all messages that
    #     it is receiving from the Fediverse.
    #     """

    # @broker.subscriber("outgoing.#", exchange="cattle_grid")
    # async def outgoing_subscriber(message: ActivityMessage) -> None:
    #     """
    #     This is a sample implementation of a subscriber
    #     for outgoing messages. cattle_grid publishes on
    #     the routing key `outgoing.ACTIVITY_TYPE` all messages that
    #     it is sending to the Fediverse.

    #     This routing key is useful if you have multiple
    #     client implementations and expect them to stay in thing.

    #         To be determined:

    #         One needs messages to synchronize clients, e.g. I have
    #         read activity with URI id on this device. These might also be send through this exchange.
    #     """

    return FastStream(
        broker,
        title="cattle_grid exchange",
        version=__version__,
        description="Exchange in cattle_grid that applications using cattle_grid as middle ware are supposed to subscribe to",
    )


def get_async_api_schema() -> Schema:
    from faststream.asyncapi import get_app_schema

    app = get_mock_faststream_app()

    return get_app_schema(app)



import click

from .cli import add_extension_commands


@click.group()
@click.pass_context
def main(ctx):
    """Tooling for managing extensions"""

    ctx.ensure_object(dict)


add_extension_commands(main)

if __name__ == "__main__":
    main()



from pydantic import BaseModel, Field, ConfigDict


class SignInData(BaseModel):
    """Used to sign into an account"""

    name: str = Field(description="Name of the account")
    password: str = Field(description="Password")


class TokenResponse(BaseModel):
    """Returns the token to be used with Bearer authentication, i.e.
    add the Header `Authorization: Bearer {token}` to the request"""

    token: str = Field(description="The token")


class LookupRequest(BaseModel):
    actor_id: str = Field(alias="actorId")
    uri: str


class LookupResponse(BaseModel):
    model_config = ConfigDict(extra="allow")

    raw: dict


class PerformRequest(BaseModel):
    """Request send to enqueue an action"""

    model_config = ConfigDict(extra="allow")

    actor: str = Field(examples=["http://actor.example/someId"])
    """The actor id, must be long to the account"""


class CreateActorRequest(BaseModel):
    """Used to create an actor for the account"""

    base_url: str = Field(
        alias="baseUrl",
        examples=["http://domain.example"],
        description="""Base url of the actor. The actor URI will be
    of the form `{baseUrl}/actor/some_secret`
    """,
    )

    handle: str | None = Field(
        None,
        examples=["alice"],
        description="""If present, an acct-uri of the form `acct:{handle}@{domain}` where domain is determined from `baseUrl` is created""",
    )

    name: str | None = Field(
        None,
        examples=["Alice"],
        description="""Internal name of the actor. Used to simplify display of the actor.""",
    )



from .util import blocklist_form_url_or_file


def test_blocklist_from_file(tmp_path):
    filename = tmp_path / "list"
    with open(filename, "w") as fp:
        fp.write(
            """blocked.test
evil.test
bad.test"""
        )

    result = blocklist_form_url_or_file(str(filename))

    assert result == {"blocked.test", "evil.test", "bad.test"}



"""Dependencies injected by fast_depends"""

import aiohttp
import logging

from typing import Annotated, Callable, Awaitable, Dict, List
from dynaconf import Dynaconf
from fast_depends import Depends
from faststream import Context
from faststream.rabbit import RabbitExchange

from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession

from cattle_grid.config.rewrite import RewriteConfiguration
from cattle_grid.model.lookup import LookupMethod
from cattle_grid.model.extension import MethodInformationModel


from .globals import (
    get_transformer,
    get_lookup,
    get_engine,
    get_method_information,
    global_container,
)

logger = logging.getLogger(__name__)


async def get_client_session():
    yield global_container.session


ClientSession = Annotated[aiohttp.ClientSession, Depends(get_client_session)]
"""The [aiohttp.ClientSession][] used by the application"""

Transformer = Annotated[Callable[[Dict], Awaitable[Dict]], Depends(get_transformer)]
"""The transformer loaded from extensions"""

LookupAnnotation = Annotated[LookupMethod, Depends(get_lookup)]
"""The lookup method loaded from extensions"""


InternalExchange = Annotated[
    RabbitExchange, Depends(global_container.get_internal_exchange)
]
"""The interal activity exchange"""


ActivityExchange = Annotated[RabbitExchange, Depends(global_container.get_exchange)]
"""The activity exchange"""

AccountExchange = Annotated[
    RabbitExchange, Depends(global_container.get_account_exchange)
]
"""The account exchange"""

SqlAsyncEngine = Annotated[AsyncEngine, Depends(get_engine)]
"""Returns the SqlAlchemy AsyncEngine"""

CorrelationId = Annotated[str, Context("message.correlation_id")]
"""The correlation id of the message"""

MethodInformation = Annotated[
    List[MethodInformationModel], Depends(get_method_information)
]
"""Returns the information about the methods that are a part of the exchange"""

SqlSessionMaker = Annotated[
    Callable[[], AsyncSession], Depends(global_container.get_session_maker)
]


async def with_sql_session(
    sql_session_maker=Depends(global_container.get_session_maker),
):
    async with sql_session_maker() as session:
        yield session


SqlSession = Annotated[AsyncSession, Depends(with_sql_session)]
"""SQL session that does not commit afterwards"""


async def with_session_commit(session: SqlSession):
    yield session
    await session.commit()


CommittingSession = Annotated[AsyncSession, Depends(with_session_commit)]
"""Session that commits the transaction"""

Config = Annotated[Dynaconf, Depends(global_container.get_config)]
"""Returns the configuration"""

RewriteRules = Annotated[
    RewriteConfiguration, Depends(global_container.get_rewrite_rules)
]
"""Rewturns the rewrite configuration"""



from faststream.rabbit import RabbitBroker

from typing import List

from cattle_grid.extensions import Extension

from cattle_grid.account.router import create_gateway_router
from cattle_grid.activity_pub.processing import create_processing_router
from cattle_grid.exchange import create_router
from cattle_grid.testing.reporter import create_reporting_router
from cattle_grid.extensions.load import (
    add_routers_to_broker as extensions_add_routers_to_broker,
    load_extensions,
    set_globals,
    collect_method_information,
)
from cattle_grid.exchange.info import exchange_method_information
from cattle_grid.dependencies.globals import global_container


def add_routers_to_broker(broker: RabbitBroker, extensions: List[Extension], settings):
    broker.include_router(create_gateway_router())
    broker.include_router(create_processing_router(global_container.internal_exchange))

    if settings.enable_reporting:
        broker.include_router(create_reporting_router())

    broker.include_router(create_router())

    extensions_add_routers_to_broker(broker, extensions)


def init_extensions(settings):
    extensions = load_extensions(settings)

    set_globals(extensions)

    global_container.method_information = (
        collect_method_information(extensions) + exchange_method_information
    )

    return extensions



from dataclasses import dataclass
from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession

from cattle_grid.account.account import account_with_name_password
from cattle_grid.database.account import Account, ActorStatus, ActorForAccount
from cattle_grid.account.permissions import allowed_base_urls
from cattle_grid.model.account import ActorInformation


def actor_to_information(actor: ActorForAccount) -> ActorInformation:
    """Transform ActorForAccount to its information ActorInformation

    ```pycon
    >>> actor = ActorForAccount(actor="http://base.example/actor", name="Alice")
    >>> actor_to_information(actor)
    ActorInformation(id='http://base.example/actor', name='Alice')

    ```
    """
    return ActorInformation(id=actor.actor, name=actor.name)


@dataclass
class AccountManager:
    """Access for managing accounts from outside cattle_grid, e.g.
    by an extension"""

    account: Account
    session: AsyncSession

    @staticmethod
    async def for_name_and_password(
        session: AsyncSession, name: str, password: str
    ) -> "AccountManager":
        """Returns an AccountManager for the given name and password"""
        account = await account_with_name_password(session, name, password)

        if account is None:
            raise ValueError("Account not found")

        return AccountManager(account=account, session=session)

    @staticmethod
    async def for_name(session: AsyncSession, name: str) -> "AccountManager":
        """Returns an AccountManager for a given name"""

        account = await session.scalar(select(Account).where(Account.name == name))

        if account is None:
            raise ValueError("Account not found")

        return AccountManager(account=account, session=session)

    def account_information(self) -> list[ActorInformation]:
        """Returns the actors belonging to the account"""

        return [
            actor_to_information(x)
            for x in self.account.actors
            if x.status == ActorStatus.active
        ]

    async def allowed_base_urls(self) -> list[str]:
        """Returns the list of base urls allowed for the account"""
        return await allowed_base_urls(self.session, self.account)



import pytest
import json
from unittest.mock import AsyncMock
from fastapi import FastAPI
from fastapi.testclient import TestClient

from bovine.crypto.helper import content_digest_sha256_rfc_9530

from .router_inbox import ap_router_inbox

from cattle_grid.testing.fixtures import *  # noqa
from cattle_grid.dependencies.globals import global_container


@pytest.fixture
def test_app():
    app = FastAPI()
    app.include_router(ap_router_inbox)
    return app


@pytest.fixture
def test_client(test_app):
    yield TestClient(test_app)


@pytest.fixture
def mock_broker(test_app):
    mock = AsyncMock()

    test_app.dependency_overrides[global_container.get_broker] = lambda: mock

    return mock


async def test_inbox_unauthorized_no_requester(
    test_client,
    actor_for_test,
):
    body = json.dumps({})
    key, val = content_digest_sha256_rfc_9530(body.encode())

    response = test_client.post(
        actor_for_test.inbox_uri,
        content=body,
        headers={
            key: val,
            "x-ap-location": actor_for_test.inbox_uri,
        },
    )
    assert response.status_code == 401


@pytest.mark.parametrize("uri_func", [lambda x: x.inbox_uri, lambda x: "/shared_inbox"])
@pytest.mark.parametrize(
    ("data", "headers"),
    [
        ({}, {}),
        ({}, {"x-cattle-grid-requester": "owner"}),
        ({"actor": "other"}, {"x-cattle-grid-requester": "owner"}),
    ],
)
async def test_inbox_unauthorized(
    uri_func,
    data,
    headers,
    test_client,
    actor_for_test,  # noqa
):
    body = json.dumps(data)
    key, val = content_digest_sha256_rfc_9530(body.encode())
    response = test_client.post(
        uri_func(actor_for_test),
        content=body,
        headers={
            key: val,
            "x-cattle-grid-requester": "owner",
            "x-ap-location": actor_for_test.inbox_uri,
        },
    )

    assert response.status_code == 401


@pytest.mark.parametrize("uri_func", [lambda x: x.inbox_uri, lambda x: "/shared_inbox"])
async def test_inbox(test_client, actor_for_test, mock_broker, uri_func):  # noqa
    body = b'{"actor": "owner", "type": "AnimalSound"}'
    key, val = content_digest_sha256_rfc_9530(body)

    response = test_client.post(
        uri_func(actor_for_test),
        content=body,
        headers={
            key: val,
            "x-cattle-grid-requester": "owner",
            "x-ap-location": actor_for_test.inbox_uri,
        },
    )

    assert response.status_code == 202

    mock_broker.publish.assert_awaited_once()


@pytest.mark.parametrize("uri_func", [lambda x: x.inbox_uri, lambda x: "/shared_inbox"])
async def test_inbox_no_digest(test_client, actor_for_test, mock_broker, uri_func):  # noqa
    response = test_client.post(
        uri_func(actor_for_test),
        json={"actor": "owner"},
        headers={
            "x-cattle-grid-requester": "owner",
            "x-ap-location": actor_for_test.inbox_uri,
        },
    )

    assert response.status_code == 400


@pytest.mark.parametrize("uri_func", [lambda x: x.inbox_uri, lambda x: "/shared_inbox"])
async def test_inbox_unprocessable(test_client, actor_for_test, uri_func):  # noqa
    body = b'{"xxxx"}'
    key, val = content_digest_sha256_rfc_9530(body)

    response = test_client.post(
        uri_func(actor_for_test),
        headers={
            key: val,
            "content-type": "text/plain",
            "x-cattle-grid-requester": "owner",
            "x-ap-location": actor_for_test.inbox_uri,
        },
        content=body,
    )

    assert response.status_code == 422


async def test_endpoint_not_found_inbox(test_client):
    response = test_client.post(
        "http://localhost/ap/inbox/not_an_actor",
        headers={
            "x-cattle-grid-requester": "owner",
            "x-ap-location": "http://localhost/ap/inbox/not_an_actor",
        },
    )

    assert response.status_code == 404



import pytest
import click

from click.testing import CliRunner

from dynaconf import LazySettings
from .key_cli import add_keys_command


@pytest.fixture
def cli(tmp_path):
    @click.group()
    @click.pass_context
    def main(ctx):
        ctx.ensure_object(dict)
        ctx.obj["config"] = LazySettings()

        ctx.obj["config_file"] = str(tmp_path / "config.toml")

    add_keys_command(main)

    return main


def test_handle_no_config():
    @click.group()
    @click.pass_context
    def main(ctx):
        ctx.ensure_object(dict)

    add_keys_command(main)
    runner = CliRunner()

    result = runner.invoke(main, ["keys", "clear"])
    assert result.exit_code == 1


def test_keys_command(cli):
    runner = CliRunner()
    result = runner.invoke(cli, ["keys", "clear"])
    assert result.exit_code == 0



import pytest

from fastapi import FastAPI
from fastapi.testclient import TestClient

from .storage import publishing_actor_for_actor_id

from .testing import *  # noqa

from . import extension


@pytest.fixture
def test_client():
    app = FastAPI()
    app.include_router(extension.api_router, prefix="/html_display")

    return TestClient(app)


def test_get_html_object(test_client):
    response = test_client.get("/html/name/2fd16a00-309b-4f3a-9d91-aa9516e59c1f")
    assert response.status_code == 404


def test_get_object_not_found(test_client):
    response = test_client.get(
        "/html_display/object/2fd16a00-309b-4f3a-9d91-aa9516e59c1f"
    )
    assert response.status_code == 404


def test_get_object(test_client, published_object):
    assert published_object

    object_id = published_object["id"]

    response = test_client.get(object_id)

    assert response.status_code == 200
    assert response.headers["content-type"] == "application/activity+json"


def test_get_object_html(test_client, published_object):
    assert published_object
    url_in_obj = published_object.get("url", [])
    url = url_in_obj[0].get("href").replace("@", "html_display/html/")

    response = test_client.get(url)

    assert response.status_code == 200
    assert response.headers["content-type"].startswith("text/html")


async def test_get_object_html_not_found(
    sql_session_for_test, test_client, actor_for_test
):
    actor = await publishing_actor_for_actor_id(
        sql_session_for_test, actor_for_test.actor_id
    )
    await sql_session_for_test.commit()

    response = test_client.get(
        f"/html_display/html/{actor.name}/2fd16a00-309b-4f3a-9d91-aa9516e59c1f"
    )

    assert response.status_code == 404


async def test_get_object_html_actor_mismatch(
    sql_session_for_test, test_client, published_object
):
    actor = await publishing_actor_for_actor_id(sql_session_for_test, "some_id")
    await sql_session_for_test.commit()

    response = test_client.get(
        f"/html_display/html/{actor.name}/" + published_object.get("id").split("/")[-1]
    )

    assert response.status_code == 404



from pydantic import Field
from cattle_grid.model.common import WithActor


class NameActorMessage(WithActor):
    """Message for renaming an actor"""

    name: str = Field(description="Name for the actor", examples=["john"])



import pytest
from unittest.mock import AsyncMock, MagicMock

from sqlalchemy.ext.asyncio import async_sessionmaker
from sqlalchemy.future import select

from cattle_grid.dependencies.globals import global_container
from cattle_grid.extensions.testing import with_test_broker_for_extension
from cattle_grid.testing.fixtures import *  # noqa
from cattle_grid.extensions.util import lifespan_for_sql_alchemy_base_class

from . import (
    extension,
    simple_storage_publish_activity,
    simple_storage_publish_object,
)
from .message_types import PublishActivity, PublishObject
from .models import StoredActivity, StoredObject, Base

activity_stub = {
    "@context": "https://www.w3.org/ns/activitystreams",
    "type": "AnimalSound",
    "actor": "http://alice.example",
    "to": ["http://bob.example"],
    "content": "moo",
}

object_stub = {
    "@context": "https://www.w3.org/ns/activitystreams",
    "type": "Note",
    "attributedTo": "http://alice.example",
    "to": ["http://bob.example"],
    "content": "moo",
}


@pytest.fixture(autouse=True)
async def sql_engine(sql_engine_for_tests):
    lifespan = lifespan_for_sql_alchemy_base_class(Base)

    async with lifespan(sql_engine_for_tests):
        yield sql_engine_for_tests


@pytest.fixture
async def sql_session(sql_engine):
    async_session = async_sessionmaker(sql_engine, expire_on_commit=False)
    yield async_session()


@pytest.fixture
async def send_message_mock():
    yield AsyncMock()


@pytest.fixture
async def test_broker(send_message_mock):
    extension.configure({"prefix": "/simple/storage/"})

    async with with_test_broker_for_extension(
        [extension], {"send_message": send_message_mock}
    ) as tbr:
        yield tbr


@pytest.mark.parametrize(
    "actor,activity",
    [
        ("http://bob.example", activity_stub),
        (
            "http://alice.example",
            {**activity_stub, "id": "http://alice.example/activity1"},
        ),
    ],
)
async def test_simple_storage_publish_activity_errors(actor, activity):
    msg = PublishActivity(
        actor=actor,
        data=activity,
    )

    with pytest.raises(ValueError):
        await simple_storage_publish_activity(
            msg, AsyncMock(), MagicMock(), AsyncMock()
        )


async def test_simple_storage_activity(sql_session, test_broker, send_message_mock):
    msg = PublishActivity(actor="http://alice.example", data=activity_stub)

    await test_broker.publish(
        msg, routing_key="publish_activity", exchange=global_container.exchange
    )

    send_message_mock.assert_awaited_once()

    args = send_message_mock.await_args[1]
    result = (await sql_session.scalars(select(StoredActivity))).all()

    assert len(result) == 1
    assert result[0].data.get("id")

    assert args["data"] == result[0].data


@pytest.mark.parametrize(
    "actor,object",
    [
        ("http://bob.example", object_stub),
        (
            "http://alice.example",
            {**object_stub, "id": "http://alice.example/activity1"},
        ),
    ],
)
async def test_simple_storage_publish_object_errors(actor, object):
    msg = PublishObject(
        actor=actor,
        data=object,
    )

    with pytest.raises(ValueError):
        await simple_storage_publish_object(
            msg,
            AsyncMock(),
            MagicMock(),
            factories=AsyncMock(),
            broker=AsyncMock(),
            correlation_id="id",
        )


async def test_simple_storage_object(
    actor_for_test, sql_session, test_broker, send_message_mock
):
    obj = object_stub.copy()
    obj["attributedTo"] = actor_for_test.actor_id
    msg = PublishObject(actor=actor_for_test.actor_id, data=obj)

    await test_broker.publish(
        msg, routing_key="publish_object", exchange=global_container.exchange
    )

    send_message_mock.assert_awaited_once()

    result = (await sql_session.scalars(select(StoredObject))).all()

    assert len(result) == 1

    assert result[0].data.get("id")



import pytest
from unittest.mock import AsyncMock

from cattle_grid.testing.fixtures import *  # noqa
from cattle_grid.model import SharedInboxMessage
from cattle_grid.database.activity_pub_actor import Following

from .shared_inbox import handle_shared_inbox_message


async def test_no_recipients_unknown_actor_nothing_happens(sql_session_for_test):
    broker = AsyncMock()
    message = SharedInboxMessage(
        data={"actor": "https://remote.test/actor", "type": "Activity"}
    )

    await handle_shared_inbox_message(
        message, session=sql_session_for_test, broker=broker
    )

    broker.publish.assert_not_awaited()


@pytest.mark.parametrize(
    "to_func",
    [
        lambda x: x.actor_id,
        lambda x: [x.actor_id],
        lambda x: [x.actor_id, "http://other.test/actor"],
    ],
)
async def test_addressed_to_actor(actor_for_test, sql_session_for_test, to_func):
    activity = {
        "actor": "http://remote.test/actor",
        "type": "Activity",
        "to": to_func(actor_for_test),
    }

    broker = AsyncMock()
    message = SharedInboxMessage(data=activity)

    await handle_shared_inbox_message(
        message, session=sql_session_for_test, broker=broker
    )

    broker.publish.assert_awaited_once()


async def test_actor_is_following(sql_session_for_test, actor_for_test):
    remote_actor = "http://remote.test/actor"
    activity = {
        "actor": remote_actor,
        "type": "Activity",
        "to": "http://remote.test/actor",
    }

    sql_session_for_test.add(
        Following(actor=actor_for_test, following=remote_actor, accepted=1, request="x")
    )
    await sql_session_for_test.commit()

    broker = AsyncMock()
    message = SharedInboxMessage(data=activity)

    await handle_shared_inbox_message(
        message, session=sql_session_for_test, broker=broker
    )

    broker.publish.assert_awaited_once()



from typing import List
from urllib.parse import urlparse

from sqlalchemy import func, select

from cattle_grid.database.activity_pub_actor import (
    Actor,
    PublicIdentifierStatus,
    PublicIdentifier,
)


def determine_preferred_username(identifiers: List[str], actor_id: str) -> str | None:
    """Determine the preferred username from the sorted identifiers.
    The result is the name of the first acct-uri whose domain matches
    the actor id.

    ```pycon
    >>> determine_preferred_username(["acct:alice@other.example",
    ...     "acct:alice@actor.example"], "http://actor.example/actor")
    'alice'

    ```
    """
    actor_domain = urlparse(actor_id).netloc
    for identifier in identifiers:
        if identifier.startswith("acct:"):
            handle, domain = identifier.removeprefix("acct:").split("@")
            if domain == actor_domain:
                return handle

    return None


def collect_identifiers_for_actor(actor: Actor) -> list[str]:
    try:
        filtered_identifiers = [
            (identifier.identifier, identifier.preference)
            for identifier in actor.identifiers
            if identifier.status
            in [PublicIdentifierStatus.verified, PublicIdentifierStatus.owned]
        ]
        return [
            x[0] for x in sorted(filtered_identifiers, key=lambda x: x[1], reverse=True)
        ]
    except Exception:
        return []


async def identifier_in_list_exists(session, identifiers: list[str]) -> bool:
    result = await session.scalar(
        select(func.count()).where(PublicIdentifier.identifier.in_(identifiers))
    )

    return result > 0



from sqlalchemy import select
from cattle_grid.dependencies.globals import global_container
from cattle_grid.extensions.examples.html_display.models import PublishingActor
from cattle_grid.extensions.examples.html_display.types import NameActorMessage
from cattle_grid.manage.actor import ActorManager
from cattle_grid.model import ActivityMessage

from . import extension
from .testing import *  # noqa


async def test_publish_html(
    test_broker, actor_for_test, mock_publish_activity, mock_publish_object
):
    obj = {
        "type": "Note",
        "to": "as:Public",
        "content": "I <3 milk!",
        "attributedTo": actor_for_test.actor_id,
    }

    await test_broker.publish(
        ActivityMessage(actor=actor_for_test.actor_id, data=obj),
        routing_key="html_display_publish_object",
        exchange=global_container.exchange,
    )

    mock_publish_activity.assert_awaited_once()
    mock_publish_object.assert_not_awaited()

    args = mock_publish_activity.await_args

    assert args[1]["actor"] == actor_for_test.actor_id


async def test_publish_html_private(
    test_broker, actor_for_test, mock_publish_activity, mock_publish_object
):
    obj = {
        "type": "Note",
        "to": ["http://remote.example/actor"],
        "content": "I <3 milk!",
        "attributedTo": actor_for_test.actor_id,
    }

    await test_broker.publish(
        ActivityMessage(actor=actor_for_test.actor_id, data=obj),
        routing_key="html_display_publish_object",
        exchange=global_container.exchange,
    )

    mock_publish_activity.assert_not_awaited()
    mock_publish_object.assert_awaited_once()


async def test_rename_actor(test_broker, actor_with_account, sql_session_for_test):
    await test_broker.publish(
        NameActorMessage(actor=actor_with_account.actor_id, name="sue"),
        routing_key="html_display_name",
        exchange=global_container.exchange,
    )

    actor = await sql_session_for_test.scalar(
        select(PublishingActor).where(
            PublishingActor.actor == actor_with_account.actor_id
        )
    )

    assert actor.name == "sue"

    manager = ActorManager(
        actor_id=actor_with_account.actor_id, session=sql_session_for_test
    )

    assert [] == await manager.groups()


async def test_rename_actor_and_add_to_group(
    test_broker, actor_with_account, sql_session_for_test
):
    old_config = extension.configuration
    extension.configure({"automatically_add_users_to_group": True})

    await test_broker.publish(
        NameActorMessage(actor=actor_with_account.actor_id, name="sue"),
        routing_key="html_display_name",
        exchange=global_container.exchange,
    )

    actor = await sql_session_for_test.scalar(
        select(PublishingActor).where(
            PublishingActor.actor == actor_with_account.actor_id
        )
    )

    assert actor.name == "sue"

    manager = ActorManager(
        actor_id=actor_with_account.actor_id, session=sql_session_for_test
    )

    assert ["html_display"] == await manager.groups()

    extension.configuration = old_config



import pytest

from uuid import uuid4
from sqlalchemy.ext.asyncio import async_sessionmaker
from sqlalchemy.future import select

from fastapi import FastAPI
from fastapi.testclient import TestClient
from cattle_grid.dependencies.globals import global_container

from cattle_grid.testing.fixtures import *  # noqa

from . import extension
from .message_types import PublishObject
from .test_handlers import object_stub, sql_engine, test_broker, send_message_mock  # noqa
from .models import StoredActivity, StoredObject


@pytest.fixture
async def stored_uuids(sql_engine, actor_for_test, test_broker):  # noqa
    obj = object_stub.copy()
    obj["attributedTo"] = actor_for_test.actor_id
    msg = PublishObject(actor=actor_for_test.actor_id, data=obj)

    async_session = async_sessionmaker(sql_engine, expire_on_commit=False)

    await test_broker.publish(
        msg, routing_key="publish_object", exchange=global_container.exchange
    )

    async with async_session() as session:
        result = (await session.scalars(select(StoredActivity))).all()

        activity_uuid = result[0].data.get("id")

        result = (await session.scalars(select(StoredObject))).all()

        object_uuid = result[0].data.get("id")

    return activity_uuid, object_uuid


@pytest.fixture
def object_ids(stored_uuids):
    _, object_uuid = stored_uuids
    return object_uuid, object_uuid.split("/")[-1]


@pytest.fixture
def activity_ids(stored_uuids):
    activity_uuid, _ = stored_uuids
    return activity_uuid, activity_uuid.split("/")[-1]


@pytest.fixture
def test_client():
    app = FastAPI()
    app.include_router(extension.api_router)

    return TestClient(app)


def test_not_found(test_client):
    response = test_client.get(
        f"/{uuid4()}",
        headers={
            "x-cattle-grid-requester": "http://bob.example",
            "x-ap-location": "location",
        },
    )

    assert response.status_code == 404


def test_get_activity(activity_ids, test_client):
    activity_id, activity_uuid = activity_ids

    response = test_client.get(
        f"/{activity_uuid}",
        headers={
            "x-cattle-grid-requester": "http://bob.example",
            "x-ap-location": activity_id,
        },
    )

    assert response.status_code == 200
    assert response.json()["type"] == "Create"
    assert response.headers["content-type"] == "application/activity+json"


def test_get_activity_unauthorized(activity_ids, test_client):
    activity_id, activity_uuid = activity_ids

    response = test_client.get(
        f"/{activity_uuid}",
        headers={
            "x-ap-location": activity_id,
        },
    )

    assert response.status_code == 401


def test_get_object(object_ids, test_client):
    object_id, object_uuid = object_ids

    response = test_client.get(
        f"/{object_uuid}",
        headers={
            "x-cattle-grid-requester": "http://bob.example",
            "x-ap-location": object_id,
        },
    )

    assert response.status_code == 200
    assert response.json()["type"] == "Note"
    assert response.headers["content-type"] == "application/activity+json"


def test_get_object_unauthorized(object_ids, test_client):
    object_id, object_uuid = object_ids

    response = test_client.get(
        f"/{object_uuid}",
        headers={
            "x-cattle-grid-requester": "http://unknown.example",
            "x-ap-location": object_id,
        },
    )

    assert response.status_code == 401


def test_get_object_wrong_location(object_ids, test_client):
    object_id, object_uuid = object_ids

    response = test_client.get(
        f"/{object_uuid}",
        headers={
            "x-cattle-grid-requester": "http://bob.example",
            "x-ap-location": "nowhere",
        },
    )

    assert response.status_code == 400



import json

from cattle_grid.model.lookup import Lookup

from .lookup import cache_lookup
from .testing import *  # noqa


async def test_cache_lookup_no_result(redis_client):
    lookup = Lookup(uri="http://actor.example/object", actor="http://actor.example")

    result = await cache_lookup(lookup, redis_client)

    assert isinstance(result, Lookup)

    assert lookup == result


async def test_cache_lookup_result(redis_client):
    uri = "http://actor.example/object"
    data = {"type": "FromCache"}
    lookup = Lookup(uri=uri, actor="http://actor.example")

    await redis_client.set(uri, json.dumps(data))

    result = await cache_lookup(lookup, redis_client)

    assert isinstance(result, Lookup)

    assert result.result == data



from .rewrite import RewriteConfiguration


def test_from_rules_none():
    config = RewriteConfiguration.from_rules(None)

    assert config.rewrite("method", ["group"]) == "method"


def test_add_new_rules():
    config = RewriteConfiguration.from_rules({})

    assert config.rewrite("method", ["group"]) == "method"

    config.add_rules("group", {"method": "changed"})

    assert config.rewrite("method", ["group"]) == "changed"


def test_add_new_rules_no_overwrite():
    config = RewriteConfiguration.from_rules({"group": {"method": "unchanged"}})

    assert config.rewrite("method", ["group"]) == "unchanged"

    config.add_rules("group", {"method": "changed"})

    assert config.rewrite("method", ["group"]) == "unchanged"



import logging
import secrets
import tomli_w
from typing import Set
from urllib.parse import urlparse

from dynaconf import Dynaconf
from pydantic import BaseModel, ValidationError, field_serializer
from functools import lru_cache

from bovine.crypto import generate_rsa_public_private_key

from .settings import get_settings

logger = logging.getLogger(__name__)


class AuthNotConfigured(Exception): ...


class AuthConfig(BaseModel):
    """Configures the Authorization layer"""

    actor_id: str
    """actor_id for the Application actor used to fetch public keys"""
    actor_acct_id: str
    """acct uri of the Application Actor used to fetch public keys"""
    public_key: str
    """Public key of the Application actor"""
    private_key: str
    """Private key of the Application actor"""

    domain_blocks: Set[str]
    """Set of blocked domains"""

    require_signature_for_activity_pub: bool = True
    """If set to true, all requests with accept type that match activitypub must be signed"""

    @field_serializer("domain_blocks")
    def serialize_domain_blocks(self, domain_blocks: Set[str], _info):
        return list(domain_blocks)


@lru_cache
def get_auth_config(settings: Dynaconf = get_settings()) -> AuthConfig:
    """Returns the configuration for authorization

    :returns:
    """
    try:
        auth = settings.get("auth")  # type:ignore
        if not auth:
            raise AuthNotConfigured("No authorization configuration found")

        data = {
            "actor_id": auth.actor_id,
            "actor_acct_id": auth.actor_acct_id,
            "public_key": auth.public_key,
            "private_key": auth.private_key,
            "domain_blocks": auth.get("domain_blocks", []),
            "require_signature_for_activity_pub": auth.get(
                "require_signature_for_activity_pub", True
            ),
        }
        return AuthConfig.model_validate(data)

    except ValidationError:
        raise AuthNotConfigured(
            "Authorization not configured or configured incorrectly"
        )


def new_auth_config(actor_id: str, username: str | None = None) -> AuthConfig:
    """Creates a new authorization configuration"""
    if not username:
        username = secrets.token_urlsafe(12)

    domain = urlparse(actor_id).netloc
    acct_uri = f"acct:{username}@{domain}"

    public_key, private_key = generate_rsa_public_private_key()

    auth_config = AuthConfig(
        actor_id=actor_id,
        actor_acct_id=acct_uri,
        public_key=public_key,
        private_key=private_key,
        domain_blocks=set(),
    )

    return auth_config


def save_auth_config(filename: str, config: AuthConfig) -> None:
    """Saves the authorization configuration to a file"""
    with open(filename, "wb") as fp:
        tomli_w.dump({"auth": config.model_dump()}, fp, multiline_strings=True)



import pytest

from cattle_grid.testing.fixtures import *  # noqa

from cattle_grid.account.account import create_account, add_permission
from cattle_grid.model.account import InformationResponse

from cattle_grid.database.account import ActorForAccount, ActorStatus

from .info import create_information_response


@pytest.fixture
async def test_admin_account(sql_session_for_test):
    account = await create_account(
        sql_session_for_test, "test_account", "test_password"
    )
    assert account
    await add_permission(sql_session_for_test, account, "admin")

    return account


async def test_create_information_response(sql_session_for_test, test_admin_account):
    response = await create_information_response(
        sql_session_for_test, test_admin_account, []
    )

    assert isinstance(response, InformationResponse)


async def test_create_information_response_actors(
    sql_session_for_test,
    test_admin_account,
):
    sql_session_for_test.add_all(
        [
            ActorForAccount(
                account=test_admin_account,
                actor="http://host.test/actor/active",
                status=ActorStatus.active,
            ),
            ActorForAccount(
                account=test_admin_account,
                actor="http://host.test/actor/deleted",
                status=ActorStatus.deleted,
            ),
        ]
    )
    await sql_session_for_test.commit()
    await sql_session_for_test.refresh(test_admin_account, attribute_names=["actors"])

    response = await create_information_response(
        sql_session_for_test, test_admin_account, []
    )

    actors = response.actors
    assert len(actors) == 1

    assert actors[0].id.endswith("active")



"""This package contains the overall router for all connection
needs to the Fediverse. This means the .well-known endpoints."""

import logging

from fastapi import APIRouter, HTTPException, Header
from fastapi.responses import JSONResponse

from typing import Annotated

from bovine.utils import webfinger_response
from bovine.types.jrd import JrdData, JrdLink
from bovine.types.nodeinfo import NodeInfo, Software
from sqlalchemy import select

from cattle_grid.dependencies.fastapi import SqlSession

from .router import ap_router
from .router_inbox import ap_router_inbox
from .router_object import ap_router_object
from cattle_grid.database.activity_pub_actor import PublicIdentifier
from cattle_grid.version import __version__

logger = logging.getLogger(__name__)

router = APIRouter(tags=["activity_pub"], prefix="/ap")
router.include_router(ap_router)
router.include_router(ap_router_inbox)
router.include_router(ap_router_object)


class JrdResponse(JSONResponse):
    """Response that ensures the content-type is
    "application/jrd+json"
    """

    media_type = "application/jrd+json"


@router.get("/")
async def main() -> str:
    return "cattle_grid ap endpoint"


@router.get("/.well-known/webfinger")
async def webfinger_responder(resource: str, session: SqlSession) -> JrdData:
    """Handles requests to .well-known/webfinger. Results are determined
    by the identifier property of [PublicIdentifier][cattle_grid.database.activity_pub_actor.PublicIdentifier] matching the resource
    parameter.

    See [RFC 7033 WebFinger](https://www.rfc-editor.org/rfc/rfc7033).
    """

    logger.info("looking up web finger for resource '%s'", resource)

    pi = await session.scalar(
        select(PublicIdentifier).where(PublicIdentifier.identifier == resource)
    )

    if not pi:
        raise HTTPException(status_code=404, detail="Item not found")

    return webfinger_response(pi.identifier, pi.actor.actor_id)


@router.get("/.well-known/nodeinfo", response_model_exclude_none=True)
async def nodeinfo_responder(x_ap_location: Annotated[str, Header()]) -> JrdData:
    return JrdData(
        links=[
            JrdLink(
                type="http://nodeinfo.diaspora.software/ns/schema/2.0",
                href=x_ap_location + "_2.0",
            )  # type: ignore
        ]
    )  # type: ignore


@router.get("/.well-known/nodeinfo_2.0", response_class=JrdResponse)
async def nodeinfo_data_responder() -> NodeInfo:
    """Returns the information according to the nodeinfo spec"""
    return NodeInfo(software=Software(name="cattle-grid", version=__version__))  # type: ignore



from sqlalchemy import select
from typing import Annotated

from fastapi import HTTPException, Depends
from fastapi.security import OAuth2PasswordBearer

# from cattle_grid.account.models import AuthenticationToken, Account
from cattle_grid.dependencies.fastapi import SqlSession
from cattle_grid.database.account import Account, AuthenticationToken


oauth2_scheme = OAuth2PasswordBearer(tokenUrl="/signin")


async def get_current_account(
    session: SqlSession, token: Annotated[str, Depends(oauth2_scheme)]
) -> Account:
    from_db = await session.scalar(
        select(AuthenticationToken).where(AuthenticationToken.token == token)
    )

    if from_db is None:
        raise HTTPException(401)

    return from_db.account


CurrentAccount = Annotated[Account, Depends(get_current_account)]
"""Annotation for the current account"""



from datetime import datetime
from enum import StrEnum, auto

from sqlalchemy import JSON, Boolean, ForeignKey, String, Text, func
from sqlalchemy.orm import Mapped, mapped_column, relationship

from .activity_pub import Base


class ActorStatus(StrEnum):
    """Represents the status of the actor"""

    active = auto()
    deleted = auto()


class PublicIdentifierStatus(StrEnum):
    """Represents the status of the public identifier"""

    unverified = auto()
    """This identifier could not be verified"""

    verified = auto()
    """This identifier was verified"""

    owned = auto()
    """This is an identifier owned by the cattle_grid instance"""


class Actor(Base):
    __tablename__ = "actor"

    id: Mapped[int] = mapped_column(primary_key=True)
    actor_id: Mapped[str] = mapped_column(String(256), unique=True)
    """The id of the actor"""
    inbox_uri: Mapped[str] = mapped_column(String(256), unique=True)
    """The uri of the inbox"""
    outbox_uri: Mapped[str] = mapped_column(String(256), unique=True)
    """The uri of the outbox"""
    following_uri: Mapped[str] = mapped_column(String(256), unique=True)
    """The uri of the following collection"""
    followers_uri: Mapped[str] = mapped_column(String(256), unique=True)
    """The uri of the followers collection"""

    preferred_username: Mapped[str] = mapped_column(String(256), nullable=True)
    """The preferred username, used as the username part of the
    acct-uri of the actor, i.e. `acct:${preferred_username}@domain`.
    See [RFC 7565 The 'acct' URI Scheme](https://www.rfc-editor.org/rfc/rfc7565.html)."""

    public_key_name: Mapped[str] = mapped_column(String(256))
    """The name given to the public key, i.e. the id will be
    `${actor_id}#${public_key_name}."""
    public_key: Mapped[str] = mapped_column(Text())
    """The public key"""

    automatically_accept_followers: Mapped[bool] = mapped_column(Boolean())
    """Set to true to indicate cattle_grid should automatically
    accept follow requests"""
    profile: Mapped[dict] = mapped_column(JSON())
    """Additional profile values"""

    status: Mapped[ActorStatus] = mapped_column(String(7))
    """Represents the status of the actor"""

    created_at: Mapped[datetime] = mapped_column(server_default=func.now())

    identifiers: Mapped[list["PublicIdentifier"]] = relationship(viewonly=True)
    followers: Mapped[list["Follower"]] = relationship(viewonly=True)
    following: Mapped[list["Following"]] = relationship(viewonly=True)
    blocking: Mapped[list["Blocking"]] = relationship(viewonly=True)


class Follower(Base):
    """The people that follow the actor"""

    __tablename__ = "follower"
    id: Mapped[int] = mapped_column(primary_key=True)
    actor_id: Mapped[int] = mapped_column(ForeignKey("actor.id", ondelete="CASCADE"))
    actor: Mapped[Actor] = relationship()

    follower: Mapped[str] = mapped_column(String(256))
    request: Mapped[str] = mapped_column(String(256))
    accepted: Mapped[bool] = mapped_column(Boolean())


class Following(Base):
    """The people the actor is following"""

    __tablename__ = "following"
    id: Mapped[int] = mapped_column(primary_key=True)
    actor_id: Mapped[int] = mapped_column(ForeignKey("actor.id", ondelete="CASCADE"))
    actor: Mapped[Actor] = relationship()

    following: Mapped[str] = mapped_column(String(256))
    request: Mapped[str] = mapped_column(String(256))
    accepted: Mapped[bool] = mapped_column(Boolean())


class Blocking(Base):
    """The people the actor is blocking"""

    __tablename__ = "blocking"
    id: Mapped[int] = mapped_column(primary_key=True)
    actor_id: Mapped[int] = mapped_column(ForeignKey("actor.id", ondelete="CASCADE"))
    actor: Mapped[Actor] = relationship()

    blocking: Mapped[str] = mapped_column(String(256))
    request: Mapped[str] = mapped_column(String(256))
    active: Mapped[bool] = mapped_column(Boolean())


class StoredActivity(Base):
    """cattle_grid generates activities under some
    circumstances (see FIXME). These will be stored
    in this table"""

    __tablename__ = "storedactivity"
    id: Mapped[str] = mapped_column(String(256), primary_key=True)
    actor_id: Mapped[int] = mapped_column(ForeignKey("actor.id", ondelete="CASCADE"))
    actor: Mapped[Actor] = relationship(lazy="joined")

    data: Mapped[dict] = mapped_column(JSON())
    published: Mapped[datetime] = mapped_column()


class PublicIdentifier(Base):
    """Public identifier"""

    __tablename__ = "publicidentifier"
    id: Mapped[int] = mapped_column(primary_key=True)
    actor_id: Mapped[int] = mapped_column(ForeignKey("actor.id", ondelete="CASCADE"))
    actor: Mapped[Actor] = relationship(lazy="joined")

    name: Mapped[str] = mapped_column(String(256))
    identifier: Mapped[str] = mapped_column(String(256), unique=True)
    preference: Mapped[int] = mapped_column(default=0)
    status: Mapped[PublicIdentifierStatus] = mapped_column(
        String(10), default=PublicIdentifierStatus.verified
    )



import logging

from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession
from faststream.rabbit import RabbitQueue
from typing import Set

from cattle_grid.activity_pub.actor.requester import update_recipients_for_actor
from cattle_grid.database.activity_pub_actor import Actor
from cattle_grid.activity_pub.activity import actor_deletes_themselves
from cattle_grid.model import ActivityMessage

logger = logging.getLogger(__name__)


def queue_for_routing_key(routing_key):
    return RabbitQueue(f"processing_{routing_key}", routing_key=routing_key)


async def update_recipients_for_collections(
    session: AsyncSession, msg: ActivityMessage, recipients: Set[str]
) -> Set[str]:
    """Updates recipients with followers and following collection."""

    db_actor = await session.scalar(select(Actor).where(Actor.actor_id == msg.actor))

    if db_actor is None:
        raise ValueError("Actor not found")

    self_delete = actor_deletes_themselves(msg.data)

    return await update_recipients_for_actor(
        session, db_actor, recipients, self_delete=self_delete
    )
